# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 12 July 2023

"""Setup and handling of Radon indexing runs of EBSD patterns in
parallel.
"""


import os
import platform
import logging
import sys
from timeit import default_timer as timer
import time

import numpy as np
import h5py

import ray
import random


from pyebsdindex import ebsd_pattern, _pyopencl_installed
from pyebsdindex._ebsd_index_single import EBSDIndexer, index_pats

if _pyopencl_installed:
    from pyebsdindex.opencl import band_detect_cl as band_detect
else:
    from pyebsdindex import band_detect as band_detect

RAYIPADDRESS = '127.0.0.1'
OSPLATFORM  = platform.system()
#if OSPLATFORM  == 'Darwin':
#    RAYIPADDRESS = '0.0.0.0'  # the localhost address does not work on macOS when on a VPN


def index_pats_distributed(
    patsin=None,
    filename=None,
    phaselist=["FCC"],
    vendor=None,
    PC=None,
    sampleTilt=70.0,
    camElev=5.3,
    bandDetectPlan=None,
    nRho=90,
    nTheta=180,
    tSigma=None,
    rSigma=None,
    rhoMaskFrac=0.1,
    nBands=9,
    patstart=0,
    npats=-1,
    chunksize=0,
    ncpu=-1,
    return_indexer_obj=False,
    ebsd_indexer_obj=None,
    keep_log=False,
    gpu_id=None,
    verbose=0
):
    """Index EBSD patterns in parallel.

    Parameters
    ----------
    patsin : numpy.ndarray, optional
        EBSD patterns in an array of shape (n points, n pattern
        rows, n pattern columns). If not given, these are read from
        ``filename``.
    filename : str, optional
        Name of file with EBSD patterns. If not given, ``patsin`` must
        be passed.
    phaselist : list of str, optional
        Options are ``"FCC"`` and ``"BCC"``. Default is ``["FCC"]``.
    vendor : str, optional
        Which vendor convention to use for the pattern center (PC) and
        the returned orientations. The available options are ``"EDAX"``
        (default), ``"BRUKER"``, ``"OXFORD"``, ``"EMSOFT"``,
        ``"KIKUCHIPY"``.
    PC : list, optional
        Pattern center (PCx, PCy, PCz) in the :attr:`indexer.vendor` or
        ``vendor`` convention. For EDAX TSL, this is (x*, y*, z*),
        defined in fractions of pattern width with respect to the lower
        left corner of the detector. If not passed, this is set to (x*,
        y*, z*) = (0.471659, 0.675044, 0.630139). If
        ``vendor="EMSOFT"``, the PC must be four numbers, the final
        number being the pixel size.
    sampleTilt : float, optional
        Sample tilt towards the detector in degrees. Default is 70
        degrees. Unused if ``ebsd_indexer_obj`` is passed.
    camElev : float, optional
        Camera elevation in degrees. Default is 5.3 degrees. Unused
        if ``ebsd_indexer_obj`` is passed.
    bandDetectPlan : pyebsdindex.band_detect.BandDetect, optional
        Collection of parameters using in band detection. Unused if
        ``ebsd_indexer_obj`` is passed.
    nRho : int, optional
        Default is 90 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    nTheta : int, optional
        Default is 180 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    tSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rhoMaskFrac : float, optional
        Default is 0.1. Unused if ``ebsd_indexer_obj`` is passed.
    nBands : int, optional
        Number of detected bands to use in triplet voting. Default
        is 9. Unused if ``ebsd_indexer_obj`` is passed.
    patstart : int, optional
        Starting index of the patterns to index. Default is ``0``.
    npats : int, optional
        Number of patterns to index. Default is ``-1``, which will
        index up to the final pattern in ``patsin``.
    chunksize : int, optional
        If not set, we will make a guess based on the resources
        available.
    ncpu : int, optional
        Number of CPUs to use. Default value is ``-1``, meaning all
        available CPUs will be used.
    return_indexer_obj : bool, optional
        Whether to return the EBSD indexer. Default is ``False``.
    ebsd_indexer_obj : EBSDIndexer, optional
        EBSD indexer. If not given, many of the above parameters must be
        passed. Otherwise, these parameters are retrieved from this
        indexer.
    keep_log : bool, optional
        Whether to keep the log. Default is ``False``.
    gpu_id : int, optional
        ID of GPU to use if :mod:`pyopencl` is installed.
    verbose : int, optional
        0 - no output (default), 1 - timings, 2 - timings and the Radon
        transform of the first pattern with detected bands highlighted.

    Returns
    -------
    indxData : numpy.ndarray
        Structured numpy array, that is
        [nphases + 1, npoints]. The data is stored for each phase used
        in indexing and the ``indxData[-1]`` layer uses the best guess
        on which is the most likely phase, based on the fit, and number
        of bands matched for each phase. Each data entry contains the
        orientation expressed as a quaternion ('quat') (using the
        convention of ``vendor`` or :attr:`indexer.vendor`), Pattern
        Quality ('pq'), Confidence Metric ('cm'), Phase ID ('phase'), Fit
        ('fit') and Number of Bands Matched ('nmatch'). There are some other
        metrics reported, but these are mostly for debugging purposes.
        The number and order of fields are not guaranteed to remain the same, but
        fields listed here are stable.
        (phase) parameter will be set to -1 for any no-solution point.
    bandData : numpy.ndarray
        Band identification data from the Radon transform. Stored
        as a structured numpy array, of dimensions [npoints, nbands].

        With fields that include:
            - id: band ID
            - max: peak max intesensity (used to calculate pattern quality)
            - maxloc: nearest integer location of the Radon peak
            - avemax: nearest neighbor average of the max peak intensity
            - aveloc: sub-pixel location of the Radon peak
            - width: a metric of the band width
            - theta: the theta value of the sub-pixel location on the Radon (lower-left origin)
            - rho: the rho value of the sub-pixel location on the Radon (lower-left origin)
            - valid: was the peak detected
            - band_match_index: index for phase number and pole number that indexed to this band
              (use :meth:`~EBSDIndexer.getmatchedpole`)

    indexer : EBSDIndexer
        EBSD indexer, returned if ``return_indexer_obj=True``.

    Notes
    -----
    Requires the ``ray[default]`` package. See the :doc:`installation
    guide </user/installation>` for details.
    """
    starttime = timer()
    pats = None
    if patsin is None:
        pdim = None
    else:
        if isinstance(patsin, ebsd_pattern.EBSDPatterns):
            pats = patsin.patterns
        if type(patsin) is np.ndarray:
            pats = patsin
        if isinstance(patsin, h5py.Dataset):
            shp = patsin.shape
            if len(shp) == 3:
                pats = patsin
            if len(shp) == 2:  # just read off disk now.
                pats = patsin[()]
                pats = pats.reshape(1, shp[0], shp[1])

        if pats is None:
            print("Unrecognized input data type")
            return
        pdim = pats.shape[-2:]

    # run a test flight to make sure all parameters are set properly before being sent off to the cluster
    if ebsd_indexer_obj is None:
        indexer = EBSDIndexer(
            filename=filename,
            phaselist=phaselist,
            vendor=vendor,
            PC=PC,
            sampleTilt=sampleTilt,
            camElev=camElev,
            bandDetectPlan=bandDetectPlan,
            nRho=nRho,
            nTheta=nTheta,
            tSigma=tSigma,
            rSigma=rSigma,
            rhoMaskFrac=rhoMaskFrac,
            nBands=nBands,
            patDim=pdim,
            gpu_id=gpu_id,
        )
    else:
        indexer = ebsd_indexer_obj

    if filename is not None:
        indexer.update_file(filename)
    else:
        indexer.update_file(patDim=pats.shape[-2:])

    # Differentiate between getting a file to index or an array.
    # Need to index one pattern to make sure the indexer object is fully
    # initiated before placing in shared memory store.
    inputmode = "memorymode"
    if pats is None:
        inputmode = "filemode"
        temp, temp2, indexer = index_pats(
            npats=1, return_indexer_obj=True, ebsd_indexer_obj=indexer
        )

    if inputmode == "filemode":
        npatsTotal = indexer.fID.nPatterns
    else:
        pshape = pats.shape
        if len(pshape) == 2:
            npatsTotal = 1
            pats = pats.reshape([1, pshape[0], pshape[1]])
        else:
            npatsTotal = pshape[0]
        temp, temp2, indexer = index_pats(
            pats[0, :, :],
            npats=1,
            return_indexer_obj=True,
            ebsd_indexer_obj=indexer,
        )

    if patstart < 0:
        patstart = npatsTotal - patstart
    if npats <= 0:
        npats = npatsTotal - patstart

    # Now set up the cluster with the indexer
    n_cpu_nodes = int(os.cpu_count())
    # int(sum([ r['Resources']['CPU'] for r in ray.nodes()]))
    if ncpu != -1:
        n_cpu_nodes = int(ncpu)


    ngpu = None
    cudagpuvis = '0'
    if gpu_id is not None:
        ngpu = np.atleast_1d(gpu_id).shape[0]

    try:
        clparam = band_detect.getopenclparam()
        if clparam is None:
            ngpu = 0
            ngpupnode = 0
        else:
            if ngpu is None:
                ngpu = len(clparam.gpu)
                gpu_id = np.arange(ngpu, dtype=int)
            cudagpuvis = ''
            for cdgpu in range(len(clparam.gpu)):
                cudagpuvis += str(cdgpu)+','

            #ngpupnode = ngpu / n_cpu_nodes
    except:
        ngpu = 0
        ngpupnode = 0

    if indexer.bandDetectPlan.useCPU == True:
        ngpu = 0


    if ngpu > 0:
        ngpupro = max(12, ngpu*8)  # number of processes that will serve data to the gpu
        #ngpupro = 8
        if n_cpu_nodes < 8:
            ngpupro = min(ngpupro,8)
        if n_cpu_nodes < 2:
            ngpupro = 2
        #if OSPLATFORM == 'Linux':
        #    ngpupro = 2

        n_cpu_per_gpu = max(min(1.0, n_cpu_nodes-ngpu), 0.5/ngpu)

        ngpuwrker = ngpupro

        ngpu_per_wrker =  1.0/ngpuwrker - 1.0e-6 # fraction of a GPU to give to each worker (band finding worker)
        ncpugpu_per_wrker = n_cpu_per_gpu/ngpuwrker - 1.0e-6 # fraction of a cpu to allocate to each gpu worker

        # amount of cpu to allocate to each cpu worker (indexing worker)
        ncpucpu_per_worker = (n_cpu_nodes - ncpugpu_per_wrker * ngpuwrker)/n_cpu_nodes


        if chunksize <= 0:
            chunksize = __optimizegpuchunk__(indexer, ngpupro, gpu_id, clparam)
    else:  # no gpus detected.
        ngpu_per_wrker = 0
        usegpu = False
        ngpu_per_wrker = 0
        ngpuwrker = n_cpu_nodes
        ncpucpu_per_worker = 0.5 - 1.0e-6
        ncpugpu_per_wrker = 0.5 - 1.0e-6
        if chunksize <= 0:
            chunksize = 1000
    ncpuwrker = n_cpu_nodes

    ray.shutdown()


    # ray.init(num_cpus=n_cpu_nodes,num_gpus=ngpu,_system_config={"maximum_gcs_destroyed_actor_cached_count": n_cpu_nodes})
    # Need to append path for installs from source ... otherwise the ray
    # workers do not know where to find the PyEBSDIndex module.

    rayclust = ray.init(
        num_cpus=int(np.round(n_cpu_nodes)),
        num_gpus=ngpu*ngpuwrker,
        #dashboard_host = RAYIPADDRESS,
        _node_ip_address=RAYIPADDRESS, #"0.0.0.0",
        runtime_env={"env_vars":
                      {"PYTHONPATH": os.path.dirname(os.path.dirname(__file__)),
                      }},
        logging_level=logging.WARNING,
    )  # Supress INFO messages from ray.
    if verbose > 1:
        print("num cpu/gpu, and number of patterns per iteration:", n_cpu_nodes, ngpu, chunksize, ngpuwrker, ncpuwrker)
    #print(rayclust)
    # Place indexer obj in shared memory store so all workers can use it - this is read only.
    remote_indexer = ray.put(indexer)
    # Get the function that will collect opencl parameters - if opencl
    # is not installed, this is None, and the program will automatically
    # fall back to CPU only calculation.
    clparamfunction = band_detect.getopenclparam
    # Set up the jobs
    njobs = (np.ceil(npats / chunksize)).astype(np.compat.long)

    p_indx_start_end = [
        [i * chunksize + patstart, (i + 1) * chunksize + patstart, chunksize]
        for i in range(njobs)
    ]
    p_indx_start_end[-1][1] = npats + patstart
    p_indx_start_end[-1][2] = p_indx_start_end[-1][1] - p_indx_start_end[-1][0]

    gpujobs = []
    cpujobs = []
    jid = 1
    for jb in p_indx_start_end:
        gpujobs.append(CPUGPUJob(jid, jb[0], jb[1]))
        jid += 1


    if njobs < ncpuwrker:
        ncpuwrker = njobs
    if njobs < ngpuwrker:
        ngpuwrker = njobs

    nPhases = len(indexer.phaseLib)
    dataout = np.zeros((nPhases + 1, npats), dtype=indexer.dataTemplate)
    banddataout = np.zeros(
        (npats, indexer.bandDetectPlan.nBands), dtype=indexer.bandDetectPlan.dataType
    )
    bandnormsout = np.zeros((npats, indexer.bandDetectPlan.nBands, 3), dtype=np.float32)


    ncpudone = 0
    ngpudone = 0
    ncpusubmit = 0
    ngpusubmit = 0
    tic0 = timer()
    ncpupatsdone = 0.0

    if keep_log is True:
        if OSPLATFORM != 'Windows':
            newline = "\n"
        else:
            newline = "\r\n"
    else:
        newline = "\r"

    # Send out the first batch
    gpuworkers = []
    gputask = []
    gtaskindex = []
    cpuworkers = []
    cputask = []
    ctaskindex = []
    npatdone = 0.0
    chunkave = 0.0

    #print(ngpuwrker, ncpugpu_per_wrker, ngpu_per_wrker)
    #print(ncpuwrker, ncpucpu_per_worker)
    #gpu_launched = 0
    #cpu_launched = 0
    while (len(gpuworkers) < ngpuwrker) and (len(gpujobs) > 0):
    #if (len(gpuworkers) < ngpuwrker) and (len(gpujobs) > 0):

        i = len(gpuworkers)
        gpuworkers.append(  # make a new Ray Actor that can call the indexer defined in shared memory.
            # These actors are read/write, thus can initialize the GPU queues
            # GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
            GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
                actorid=i, clparammodule=clparamfunction, gpu_id=gpu_id, cudavis=cudagpuvis
            )
        )
        gjob = gpujobs.pop(0)
        if inputmode == "filemode":
            gputask.append(
                gpuworkers[i].findbands.remote(gjob,
                                               pats=None,
                                               indexer=remote_indexer
                                               )
            )
        else:
            gputask.append(
                gpuworkers[i].findbands.remote(gjob,
                                               pats=pats[gjob.pstart:gjob.pend, :, :],
                                               indexer=remote_indexer,
                                               )
            )
        gtaskindex.append(gjob)

        #gpu_launched += 1

    gpuwrker_cycles = 0
    cpuwrker_cycles = 0

    while ncpudone < njobs:


        # initiate the CPU workers.
        #print(len(gpuworkers), len(gputask))
        #for i in range(ncpuwrker):
        if (len(cpuworkers) < ncpuwrker) and ((njobs - ncpudone) > len(cpuworkers)):
            i = len(cpuworkers)
            cpuworkers.append(  # make a new Ray Actor that can call the indexer defined in shared memory.
                # These actors are read/write, thus can initialize the GPU queues
                CPUWorker.options(num_cpus=ncpucpu_per_worker).remote(i))
                #CPUWorker.options(num_cpus=1.0, num_gpus=0).remote(i))
            cputask.append(cpuworkers[i].indexpoles.remote(None, None, None,indexer=remote_indexer))
            ctaskindex.append(None)
            #cpu_launched += 1
        #print(len(cpuworkers))


        # check if GPU is working
        if ngpudone < njobs: # check if gpu is done

            gpuwrker_cycles +=1

            donewrker, busy = ray.wait(gputask,num_returns=len(gputask),  timeout=0.1)

            #print(len(donewrker), nret)
            #print()
            #nret = max(1,len(gputask)//2)
            #donewrker, busy = ray.wait(gputask, num_returns=nret, timeout=1.0)
            #if len(wrker) > 0:  # trying to catch a hung worker.  Rare, but it happens
            #print(len(donewrker))
            #else:
                #print('hung gpu process')
                #jid = gputask.index(busy[0])
                #wrker.append(busy[0])
                #ray.kill(gputask[jid])
            for wrker in donewrker:
                
                gpuwrker_cycles = 0
                jid = gputask.index(wrker)
                try:
                    message, (banddata, bandnorm, gjob) = ray.get(wrker)

                    if message == 'Done':
                        banddataout[gjob.pstart - patstart: gjob.pend - patstart, :] = banddata
                        bandnormsout[gjob.pstart - patstart: gjob.pend - patstart, :,:] = bandnorm

                        cpujobs.append(CPUGPUJob(gjob.jobid, gjob.pstart, gjob.pend, extime=gjob.extime))
                        ngpudone += 1

                    if len(gpujobs) > 0: # still more gpu work to do
                        gjob = gpujobs.pop(0)
                        if inputmode == "filemode":
                            gputask[jid] = gpuworkers[jid].findbands.remote(gjob,
                                    pats=None,
                                    indexer=remote_indexer
                            )
                        else:
                            gputask[jid] = gpuworkers[jid].findbands.remote(gjob,
                                pats=pats[gjob.pstart:gjob.pend, :, :],
                                indexer=remote_indexer,
                           )
                        gtaskindex[jid] = gjob
                        ngpusubmit += 1


                    else: # no more gpu tasks to submit
                        #del gpuworkers[jid]
                        #ray.kill(gpuworkers[jid])
                        del gpuworkers[jid]
                        del gputask[jid]
                        del gtaskindex[jid]
                        time.sleep(0.001)
                    #else:
                    #    raise Exception("Error in GPU processing patterns: ", gtaskindex[jid].pstart, gtaskindex[jid].pend)



                except:
                    gjob = gtaskindex[jid]
                    print('A GPU death has occured', gjob.pstart, gjob.pend)
                    del gpuworkers[jid]
                    del gputask[jid]
                    del gtaskindex[jid]
                    gpujobs.append(gjob)
                    if len(gpuworkers) == 0:
                        gpuworkers.append(  # make a new Ray Actor that can call the indexer defined in shared memory.
                            # These actors are read/write, thus can initialize the GPU queues
                            # GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
                            GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
                                actorid=0, clparammodule=clparamfunction, gpu_id=gpu_id, cudavis=cudagpuvis
                            )
                        )
                        gjob = gpujobs.pop(0)
                        if inputmode == "filemode":
                            gputask.append(
                                gpuworkers[0].findbands.remote(gjob, pats=None,
                                   indexer=remote_indexer
                                )
                            )
                        else:
                            gputask.append(
                                gpuworkers[0].findbands.remote(gjob,
                                   pats=pats[gjob.pstart:gjob.pend, :, :],
                                   indexer=remote_indexer,
                                )
                            )
                        gtaskindex.append(gjob)
            # toc = timer()

            if gpuwrker_cycles > 100: # a gpu worker got stuck -- see if I can unstick it.
                wrker = busy[0]
                gpuwrker_cycles = 0
                jid = gputask.index(wrker)
                gjob = gtaskindex[jid]
                print('A GPU death has occured. Attempting to restart.', gjob.pstart, gjob.pend)
                ray.kill(gpuworkers[jid])
                del gpuworkers[jid]
                del gputask[jid]
                del gtaskindex[jid]
                gpujobs.append(gjob)
                if len(gpuworkers) == 0:
                    gpuworkers.append(  # make a new Ray Actor that can call the indexer defined in shared memory.
                        # These actors are read/write, thus can initialize the GPU queues
                        # GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
                        GPUWorker.options(num_cpus=ncpugpu_per_wrker, num_gpus=ngpu_per_wrker).remote(
                            actorid=0, clparammodule=clparamfunction, gpu_id=gpu_id, cudavis=cudagpuvis
                        )
                    )
                    if inputmode == "filemode":
                        gputask.append(
                            gpuworkers[0].findbands.remote(gjob,
                                                           pats=None,
                                                           indexer=remote_indexer
                                                           )
                        )
                    else:
                        gputask.append(
                            gpuworkers[0].findbands.remote(gjob,
                                                           pats=pats[gjob.pstart:gjob.pend, :, :],
                                                           indexer=remote_indexer,
                                                           )
                        )
                    gtaskindex.append(gjob)

            if (ngpudone >= njobs) and (verbose > 1 ):
                print('\nGPU Done')
        if ncpudone < njobs:
            donewrker, busy = ray.wait(cputask, num_returns = len(cputask),  timeout=0.01)
            #donewrker, busy = ray.wait(cputask, num_returns=1, timeout=1.0)
            for wrker in donewrker:

                jid = cputask.index(wrker)
                try:
                    message, (indexdata,bnddata, cjob) = ray.get(wrker)
                    if message == 'Done':
                        dataout[:, cjob.pstart - patstart: cjob.pend - patstart] = indexdata
                        banddataout[cjob.pstart - patstart: cjob.pend - patstart, :] = bnddata
                        ncpudone += 1
                        chunkave += cjob.rate
                        npatdone += cjob.npat
                        currenttime = timer() - starttime
                        #print(cjob.rate * n_cpu_nodes)
                        if verbose > 0:
                            #print()
                            print(
                                "Completed: ",
                                str(cjob.pstart),
                                " -- ",
                                str(cjob.pend),
                                "  PPS:",
                                #"{:.0f}".format(cjob.rate*ncpuwrker)
                                #+ ";"
                                #+ "{:.0f}".format(chunkave / ncpudone * ncpuwrker)
                                #+ ";" +
                                "{:.0f}".format(npatdone/currenttime),
                                "  ",
                                "{:.0f}".format((ncpudone / njobs) * 100) + "%",
                                "{:.0f};".format(currenttime)
                                + "{:.0f}".format((njobs - ncpudone) / ncpudone * currenttime)
                                + " running;remaining(s)",
                                end=newline,
                            )
                        #time.sleep(0.001)
                    if message != 'Error':
                        if ncpudone == njobs:
                            #cpuworkers[jid] = None
                            #cputask[jid] = None
                            #ctaskindex[jid] = None
                            #ray.kill(cpuworkers[jid])
                            del cpuworkers[jid]
                            del cputask[jid]
                            del ctaskindex[jid]
                            time.sleep(0.1)
                        elif len(cpujobs) > 0:
                            cjob = cpujobs.pop(0)
                            if cjob is not None:
                                banddata = banddataout[cjob.pstart - patstart: cjob.pend - patstart, :]
                                bandnorm = bandnormsout[cjob.pstart - patstart: cjob.pend - patstart, :, :]
                                cputask[jid] = cpuworkers[jid].indexpoles.remote(cjob,banddata, bandnorm, indexer=remote_indexer)
                                ctaskindex[jid] = cjob
                            else:
                                cputask[jid] = cpuworkers[jid].indexpoles.remote(None, None, None)
                                ctaskindex[jid] = None

                        else: # there should be more to do, but waiting for work.
                            cputask[jid] = cpuworkers[jid].indexpoles.remote(None, None, None)
                            ctaskindex[jid] = None

                    else:
                        raise Exception("Error in indexing bands")#, ctaskindex[jid].pstart,ctaskindex[jid].pend)


                except Exception as e:
                    print(e)
                    cjob = ctaskindex[jid]
                    print('A CPU death has occured')
                    ray.kill(cpuworkers[jid])
                    del cpuworkers[jid]
                    del cputask[jid]
                    del ctaskindex[jid]
                    cpujobs.append(cjob)
                    if len(cpuworkers) == 0:
                        cpuworkers.append(  # make a new Ray Actor that can call the indexer defined in shared memory.
                            # These actors are read/write, thus can initialize the GPU queues
                            CPUWorker.options(num_cpus=1, num_gpus=0).remote(0))
                        cputask.append(cpuworkers[0].indexpoles.remote(None, None, None))
                        ctaskindex.append(None)
    if verbose > 0:
        print('\n...')
    ray.shutdown()

    if return_indexer_obj:
        return dataout, banddataout, indexer
    else:
        return dataout, banddataout


def __optimizegpuchunk__(indexer, ngpupro, gpu_id, clparam):


    gpulist = []
    # test for GPU presence
    if clparam is None:
        return 1000

    if clparam.ngpu == 0:
        return 1000

    if gpu_id is None:
        for g in clparam.gpu:
            gpulist.append(g)
    else:
        temp = np.atleast_1d(gpu_id)

        for g in temp:
            gpulist.append(clparam.gpu[g])
    ngpu = len(gpulist)

    if ngpu == 0:
        return 1000

    gmem = 1e99
    for g in gpulist:

        if g.global_mem_size < gmem:
            gmem = g.global_mem_size

    gmem = 0.8*gmem # Build a margin in.
    #print('Global Mem:', gmem)
    #ncpu_per_gpu = max(1, np.ceil(n_cpu_nodes/ngpu))

    #print('Ncpu/gpu:', ncpu_per_gpu)
    patdim = indexer.bandDetectPlan.patDim
    rdndim = np.array([indexer.bandDetectPlan.nTheta+2*indexer.bandDetectPlan.padding[1],
                       indexer.bandDetectPlan.nRho+2*indexer.bandDetectPlan.padding[0]])
    memperpat = 4.0*float(patdim[0] * patdim[1] + 6.0 * rdndim[0] * rdndim[1])# rough estimate

    #print('Mem/pat:', memperpat)
    chunkguess = ngpu*(float(gmem)/float(ngpupro)) / memperpat

    #print('chunkguess:', chunkguess)


    #print('cheatguess:', chunkguess)
    chunk = int(max(2, np.floor(chunkguess/16))*16) # ideally should be a multiple of 16
    #print('chunk:', chunk)
    #check for powers of two: it is adventageous to be near those.
    twocheck = np.log2(float(chunk))
    if np.abs((twocheck) - np.round(twocheck)) < 0.2:
        chunk = int(2**int(np.round(twocheck)))
        if OSPLATFORM == 'Darwin': # I don't know why, but AMD/Intel macOS does not like powers of two.
            chunk = max(48, chunk-16)

    # finally - I am unsure how to check for integrated graphics that report system memory, so I am going
    # throw an arbitrary cap on this:
    chunk = min(2032, chunk)

    return chunk


@ray.remote(num_cpus=1, num_gpus=1)
class GPUWorker:
    def __init__(self, actorid=0, clparammodule=None, gpu_id=None, cudavis = '0'):
        # sys.path.append(path.dirname(path.dirname(__file__)))  # do this to help Ray find the program files
        # import openclparam # do this to help Ray find the program files
        # device, context, queue, program, mf
        # self.dataout = None
        # self.indxstart = None
        # self.indxend = None
        # self.rate = None
        os.environ["CUDA_VISIBLE_DEVICES"] = cudavis
        self.actorID = actorid
        self.openCLParams = None
        self.useGPU = False
        if clparammodule is not None:
            try:
                if (
                    sys.platform != "darwin"
                ):  # linux with NVIDIA (unsure if it is the os or GPU type) is slow to make a context
                    self.openCLParams = clparammodule()
                else:  # MacOS handles GPU memory conflicts much better when the context is destroyed between each
                    # run, and has very low overhead for making the context.
                    # pass
                    self.openCLParams = clparammodule()
                    # self.openCLParams.gpu_id = 0
                    # self.openCLParams.gpu_id = 1
                    # self.openCLParams.gpu_id = self.actorID % self.openCLParams.ngpu
                if gpu_id is None:
                    gpu_id = np.arange(self.openCLParams.ngpu)
                gpu_list = np.atleast_1d(gpu_id)
                ngpu = gpu_list.shape[0]
                #print(self.actorID, ngpu)
                self.openCLParams.gpu_id = gpu_list[self.actorID % ngpu]
                self.openCLParams.get_context()
                #self.openCLParams.get_queue()
                self.useGPU = True
            except:
                self.openCLParams = None

    def findbands(self, gpujob, pats=None, xyloc=None, PC = None, indexer=None):

        if gpujob is None:
            #time.sleep(0.001)
            return 'Bored', (None, None, None)
        try:
            # print(type(self.openCLParams.ctx))
            gpujob._starttime()
            #time.sleep(random.uniform(0, 1.0))
            if PC is None:
                PC = indexer.PC

            if self.openCLParams is not None:
                self.openCLParams.get_queue()

            pats, xyloc = indexer._getpats(
                patsin=pats,
                patstart=gpujob.pstart,
                npats=gpujob.npat,
                xyloc=xyloc)

            npoints = pats.shape[0]

            banddata, bandnorm = indexer._detectbands(pats, PC,
                                                   xyloc=xyloc,
                                                   clparams=self.openCLParams,
                                                   chunksize=-1)
            if self.openCLParams is not None:
                self.openCLParams.queue.finish()
                self.openCLParams.queue = None
            gpujob._endtime()
            #print(gpujob.endtime - gpujob.starttime )

            return 'Done', (banddata, bandnorm, gpujob)
        except:
            gpujob.rate = None
            return "Error", (None, None, gpujob)


@ray.remote(num_cpus=1, num_gpus=0)
class CPUWorker:
    def __init__(self, actorid=0):
        self.actorID = actorid

    def indexpoles(self, cpujob, banddata, bandnorm, indexer=None):
        if cpujob is None:
            #time.sleep(0.01)
            return 'Bored', (None, None, None)
        try:
            # print(type(self.openCLParams.ctx))

            cpujob._starttime()

            indxData, banddata = indexer._indexbandsphase(banddata, bandnorm, verbose=0)

            cpujob._endtime()
            return "Done", (indxData,banddata, cpujob)
        except Exception as e:
            print(e)
            cpujob.rate = None
            return "Error", (None,None, cpujob)


class CPUGPUJob:
    def __init__(self,jobid, pstart, pend, extime=0.0):
        self.jobid = jobid
        self.pstart = pstart
        self.pend = pend
        self.npat = pend - pstart
        self.starttime = 0.0
        self.endtime = 0.0
        self.extime = extime
        self.rate = 0.0
    def _starttime(self):
        self.starttime = timer()
    def _endtime(self):
        self.endtime = timer()
        self.extime += self.endtime - self.starttime
        self.rate = self.npat/(self.extime + 1e-12)




# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a
# full list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import datetime
import inspect
import os
from os.path import relpath, dirname
import re
import sys

from pyebsdindex import __author__, __file__, __version__


# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))

project = "PyEBSDIndex"
copyright = f"{datetime.now().year}, {__author__}"
author = __author__
version = __version__
release = __version__

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_gallery.load_style",
]

# Create links to references within the documentation to these packages
intersphinx_mapping = {
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]

# HTML theming: pydata-sphinx-theme
# https://pydata-sphinx-theme.readthedocs.io
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "logo": {
        "alt_text": project,
    },
    "github_url": "https://github.com/USNavalResearchLaboratory/PyEBSDIndex",
    "header_links_before_dropdown": 6,
    "navigation_with_keys": False,
    "show_toc_level": 2,
    "use_edit_page_button": True,
}
html_context = {
    "github_user": "USNavalResearchLaboratory",
    "github_repo": "PyEBSDIndex",
    "github_version": "main",
    "doc_path": "doc",
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Syntax highlighting
pygments_style = "friendly"

# -- nbsphinx configuration --------------------------------------------
nbsphinx_execute = "auto"
nbsphinx_execute_arguments = [
    "--InlineBackend.rc=figure.facecolor='w'",
    "--InlineBackend.rc=font.size=15",
]


def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object.
    This is taken from SciPy's ``conf.py``:
    https://github.com/scipy/scipy/blob/main/doc/source/conf.py.
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    if lineno:
        linespec = "#L%d-L%d" % (lineno, lineno + len(source) - 1)
    else:
        linespec = ""

    startdir = os.path.abspath(os.path.join(dirname(__file__), ".."))
    fn = relpath(fn, start=startdir).replace(os.path.sep, "/")

    if fn.startswith("pyebsdindex/"):
        m = re.match(r"^.*dev0\+([a-f\d]+)$", __version__)
        pre_link = "https://github.com/USNavalResearchLaboratory/PyEBSDIndex/blob/"
        if m:
            return pre_link + "%s/%s%s" % (m.group(1), fn, linespec)
        elif "dev" in __version__:
            return pre_link + "main/%s%s" % (fn, linespec)
        else:
            return pre_link + "v%s/%s%s" % (__version__, fn, linespec)
    else:
        return None


# -- Copy button customization (taken from PyVista)
# Exclude traditional Python prompts from the copied code
copybutton_prompt_text = r">>> ?|\.\.\. "
copybutton_prompt_is_regexp = True

# -- sphinx.ext.autodoc
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
autosummary_ignore_module_all = False
autosummary_imported_members = True
autodoc_typehints_format = "short"
autodoc_default_options = {
    "show-inheritance": True,
}

# -- numpydoc
# https://numpydoc.readthedocs.io
numpydoc_show_class_members = False
numpydoc_use_plots = True
numpydoc_xref_param_type = True
numpydoc_validation_checks = {
    "all",   # All but the following:
    "ES01",  # Not all docstrings need an extended summary
    "EX01",  # Examples: Will eventually enforce
    "GL01",  # Contradicts numpydoc examples
    "GL02",  # Appears to be broken?
    "GL07",  # Appears to be broken?
    "GL08",  # Methods can be documented in super class
    "PR01",  # Parameters can be documented in super class
    "PR02",  # Properties with setters might have docstrings w/"Returns"
    "PR04",  # Doesn't seem to work with type hints?
    "RT01",  # Abstract classes might not have return sections
    "SA01",  # Not all docstrings need a "See Also"
    "SA04",  # "See Also" section does not need descriptions
    "SS06",  # Not possible to make all summaries one line
    "YD01",  # Yields: No plan to enforce
}



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 22 May 2024

# For more info see
# Patrick T. Brewick, Stuart I. Wright, David J. Rowenhorst. Ultramicroscopy, 200:50–61, May 2019.



from pathlib import Path
from timeit import default_timer as timer

import numba
import numpy as np
import scipy.optimize as opt

from pyebsdindex import ebsd_pattern

#from os import environ
#environ["NUMBA_CACHE_DIR"] = str(tempdir)


__all__ = [
    "NLPAR",
]


class NLPAR:
  def __init__(self, filename=None,  lam=0.7, searchradius=3,dthresh=0.0, nrows = None, ncols = None, **kwargs):
    self.lam = lam
    self.searchradius = searchradius
    self.dthresh = dthresh
    self.filepath = None
    self.hdfdatapath = None
    self.filepathout = None
    self.hdfdatapathout = None
    #self.patternfile = None
    #self.patternfileout = None
    self.setfile(filename)
    self.mask = None
    self.sigma = None
    self.sigmann = 1
    self.nrows = None
    self.ncols = None
    if nrows is not None:
      self.nrows = nrows
    if ncols is not None:
      self.ncols = ncols


  def setfile(self,filepath=None):
    self.filepath = None
    self.hdfdatapath = None
    pathtemp = np.atleast_1d(filepath)
    fpath = pathtemp[0]
    hdf5path = None
    if pathtemp.size > 1:
      hdf5path = pathtemp[1]
    if fpath is not None:
      self.filepath = Path(fpath)
      self.hdfdatapath = hdf5path

  def setoutfile(self, patternfile, filepath=None):
    """Set the output file.

    Parameters
    ----------
    patternfile
      Input pattern file object from ebsd_pattern.
    filepath
      String.

    Notes
    -----
    In the future I want to be able to specify the HDF5 data path to
    store the output data, but that is proving to be a bit of a mess.
    For now, a copy of the original HDF5 is made, and the NLPAR patterns
    will be overwritten on top of the originals.
    """
    self.filepathout = None
    self.hdfdatapathout = None
    pathtemp = np.atleast_1d(filepath)
    fpath = pathtemp[0]
    hdf5path = None
    #if pathtemp.size > 1:
    #  hdf5path = pathtemp[1]
    #print(fpath, hdf5path)
    if fpath is not None: # the user has set an output file path.
      self.filepathout = Path(fpath).expanduser().resolve()
      self.hdfdatapathout =  hdf5path
      if patternfile.filetype != 'HDF5': #check that the input and output are not the same.
        pathok = self.filepathout.exists()
        if pathok:
          pathok = not self.filepathout.samefile(patternfile.filepath)
          if not pathok:
            raise ValueError('Error: File input and output are exactly the same.')
            return

        patternfile.copy_file([self.filepathout,self.hdfdatapathout], empty_data=True)
        return  # fpath and (maybe) hdf5 path were set manually.
      else: # this is a hdf5 file
        if self.hdfdatapathout is None:
          patternfile.copy_file(self.filepathout, empty_data=True)
          self.hdfdatapathout = patternfile.h5patdatpth
          return
        else:
          patternfile.copy_file([self.filepathout, self.hdfdatapathout], empty_data=True)
          return

    if patternfile is not None: # the user has set no path.
      hdf5path = None
      
      if patternfile.filetype in ['UP', 'EBSP']:
        p = Path(patternfile.filepath)
        appnd = "_NLPAR_l{:1.2f}".format(self.lam) + "sr{:d}".format(self.searchradius)
        newfilepath = str(p.parent / Path(p.stem + appnd + p.suffix))
        patternfile.copy_file(newfilepath,empty_data=True)

      if patternfile.filetype == 'HDF5':
        hdf5path_tmp = str(patternfile.h5patdatpth).split('/')
        if hdf5path_tmp[0] == '':
          hdf5path_org =  hdf5path_tmp[1]
        else:
          hdf5path_org = hdf5path_tmp[0]
        p = Path(patternfile.filepath)
        appnd = "_NLPAR_l{:1.2f}".format(self.lam) + "sr{:d}".format(self.searchradius)
        hdf5path = hdf5path_org+appnd
        newfilepath = str(p.parent / Path(p.stem + appnd + p.suffix))
        #patternfile.copy_file([newfilepath, hdf5path_org], newh5path=hdf5path)
        patternfile.copy_file([newfilepath], empty_data=True)
        hdf5path = patternfile.h5patdatpth

      self.filepathout = newfilepath
      self.hdfdatapathout = hdf5path
      return

  def getinfileobj(self):
    if self.filepath is not None:
      fID = ebsd_pattern.get_pattern_file_obj([self.filepath, self.hdfdatapath])
      if (fID.nRows is not None):
        if (self.nrows is None):
          self.nrows = fID.nRows
        else:
          fID.nRows = self.nrows

      if (fID.nCols is not None):
        if (self.ncols is None):
          self.ncols = fID.nCols
        else:
          fID.nCols = self.ncols

      return fID

    else:
      return None

  def getoutfileobj(self):
    if self.filepathout is not None:
      fID = ebsd_pattern.get_pattern_file_obj([self.filepathout, self.hdfdatapathout])
      if self.nrows is not None:
        fID.nRows = self.nrows
      else:
        self.nrows = fID.nRows

      if self.ncols is not None:
        fID.nCols = self.ncols
      else:
        self.ncols = fID.nCols
      return fID
    else:
      return None

  def opt_lambda(self,chunksize=0,saturation_protect=True,automask=True, backsub = False,
                target_weights=(0.5, 0.34, 0.25), dthresh=0.0, autoupdate=True, verbose = 2, **kwargs):

    target_weights = np.asarray(target_weights)

    def loptfunc(lam,d2,tw,dthresh):
      temp = (d2 > dthresh).choose(dthresh, d2)
      dw = np.exp(-(temp) / lam ** 2)
      w = np.sum(dw, axis=2) + 1e-12

      metric = np.mean(np.abs(tw - 1.0 / w))
      return metric


    @numba.njit(fastmath=True, cache=True,parallel=True)
    def d2norm(d2, n2, dij, sigma):
      shp = d2.shape
      s2 = sigma**2
      for j in numba.prange(shp[0]):
        for i in range(shp[1]):
          for q in range(shp[2]):
            if n2[j,i,q] > 0:
              ii = dij[j,i,q,1]
              jj = dij[j,i,q,0]
              s2_12 = (s2[j,i]+s2[jj,ii])
              d2[j,i,q] -= n2[j,i,q] * s2_12
              d2[j,i,q] /= s2_12*np.sqrt(2.0*n2[j,i,q])

    patternfile = self.getinfileobj()
    patternfile.read_header()
    nrows = np.uint64(self.nrows) #np.uint64(patternfile.nRows)
    ncols = np.uint64(self.ncols) #np.uint64(patternfile.nCols)

    pwidth = np.uint64(patternfile.patternW)
    pheight = np.uint64(patternfile.patternH)
    phw = pheight * pwidth

    nn = 1
    if chunksize <= 0:
      chunksize = np.maximum(1, np.round(1e9 / phw / ncols) ) # keeps the chunk at about 4GB
      chunksize = np.minimum(nrows,chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)

    nn = np.uint64(nn)

    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth),dtype=np.uint8)

    self.mask = (self.mask).astype(np.uint8)
    indices = np.asarray((self.mask.flatten().nonzero())[0],np.uint64)

    sigma = np.zeros((nrows,ncols),dtype=np.float32)+1e24
    colstartcount = np.asarray([0,ncols],dtype=np.int64)


    dthresh = np.float32(dthresh)
    lamopt_values = []
    
    for j in range(0,nrows,chunksize):

      if verbose >= 2:
        print("begin row: ", j, "/", nrows, sep='', end='\r')
      #print('Block',j)

      #rowstartread = np.int64(max(0,j - nn))
      rowstartread = np.int64(j)
      rowend = min(j + chunksize + nn,nrows)
      rowcountread = np.int64(rowend - rowstartread)
      data, xyloc = patternfile.read_data(patStartCount=[[0,rowstartread],[ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shp = data.shape

      if backsub is True:
        data = self.backsub(data)
        #back = np.mean(data, axis=0)
        #back -= np.mean(back)
        #data -= back
      data = data.reshape(shp[0], phw)

      rowstartcount = np.asarray([0,rowcountread],dtype=np.int64)
      sigchunk, (d2,n2, dij) = self.sigma_numba(data,nn,rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)
      tmp = (sigma[j:j + rowstartcount[1],:] < sigchunk).choose( sigchunk, sigma[j:j + rowstartcount[1],:])
      sigma[j:j + rowstartcount[1],:] = tmp


      d2norm(d2, n2, dij, sigchunk)


      lamopt_values_chnk = []
      for tw in target_weights:
        lam = 1.0
        lambopt1 = opt.minimize(loptfunc,lam,args=(d2,tw,dthresh),method='Nelder-Mead',
                                bounds = [[0.001, 10.0]],options={'fatol': 0.0001})
        lamopt_values_chnk.append(lambopt1['x'])


      lamopt_values.append(lamopt_values_chnk)

    if verbose >= 2:
      print('', end='')
    lamopt_values = np.asarray(lamopt_values)
    if verbose >=1:
      print("Range of lambda values: ", np.mean(lamopt_values, axis = 0).flatten())
      print("Optimal Choice: ", np.median(np.mean(lamopt_values, axis = 0)))

    if autoupdate == True:
      self.lam = np.median(np.mean(lamopt_values, axis = 0))
    if self.sigma is None:
      self.sigma = sigma
    return np.mean(lamopt_values, axis = 0).flatten()

  def calcnlpar(self, chunksize=0, searchradius=None, lam = None, dthresh = None, saturation_protect=True, automask=True,
               filename=None, fileout=None, reset_sigma=False, backsub = False, rescale = False,verbose=2,
                **kwargs):

    if lam is not None:
      self.lam = lam

    if dthresh is not None:
      self.dthresh = dthresh

    if searchradius is not None:
      self.searchradius = searchradius

    lam = np.float32(self.lam)
    dthresh = np.float32(self.dthresh)
    sr = np.int64(self.searchradius)

    if filename is not None:
      self.setfile(filepath=filename)




    patternfile = self.getinfileobj()

    #if filepathout is not None:
    self.setoutfile(patternfile, filepath=fileout)

    patternfileout = self.getoutfileobj()


    nrows = np.int64(self.nrows)#np.int64(patternfile.nRows)
    ncols = np.int64(self.ncols)#np.int64(patternfile.nCols)
    if patternfileout.nCols is None:
      patternfileout.nCols = ncols
    if patternfileout.nRows is None:
      patternfileout.nRows = nrows


    pwidth = np.int64(patternfile.patternW)
    pheight = np.int64(patternfile.patternH)
    phw = pheight*pwidth

    if chunksize <= 0:
      chunksize = np.maximum(1, np.round(1e9 / phw / ncols) ) # keeps the chunk at about 8GB
      chunksize = np.minimum(nrows, chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)



    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth), dtype=np.uint8)

    indices = np.asarray( (self.mask.flatten().nonzero())[0], np.uint64)
    calcsigma = False
    if self.sigma is None:
      calcsigma = True
      self.sigma = np.zeros((nrows, ncols), dtype=np.float32)+1e24

    if reset_sigma:
      self.sigma = None

    if (np.asarray(self.sigma).size == 1) and (self.sigma is not None):
      tmp = np.asarray(self.sigma)[0]
      self.sigma =  np.zeros((nrows, ncols), dtype=np.float32)+tmp
      calcsigma = False

    shpsigma = np.asarray(self.sigma).shape
    if (shpsigma[0] != nrows) and (shpsigma[1] != ncols):
      self.sigma = np.zeros((nrows,ncols),dtype=np.float32) + 1e24
      calcsigma = True

    sigma = np.asarray(self.sigma)

    scalemethod = 'clip'
    if rescale == True:
      if np.issubdtype(patternfileout.filedatatype, np.integer):
        mxval = np.iinfo(patternfileout.filedatatype).max
        scalemethod = 'fullscale'
      else: # not int, so no rescale.
        rescale = False

    nthreadpos = numba.get_num_threads()
    #numba.set_num_threads(18)
    #numba.set_num_threads(18)
    colstartcount = np.asarray([0,ncols],dtype=np.int64)
    if verbose >= 1:
      print("lambda:", self.lam, "search radius:", self.searchradius, "dthresh:", self.dthresh)

    for j in range(0,nrows,chunksize):
      #print('Row start', j)
      if verbose >= 2:
        print("begin row: ", j, "/", nrows, sep='', end='\r')

      rowstartread = np.int64(max(0, j-sr))
      rowend = min(j + chunksize+sr,nrows)

      if (rowend - rowstartread) < (2*sr+1):
        rowstartread = np.int64(max(0, rowend - (2*sr+1)))
      rowcountread = np.int64(rowend-rowstartread)
      data, xyloc = patternfile.read_data(patStartCount = [[0,rowstartread], [ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shpdata = data.shape

      if backsub is True:
        data = self.backsub(data)


      data = data.reshape(shpdata[0], phw)

      rowstartcount = np.asarray([0,rowcountread],dtype=np.int64)
      if calcsigma is True:
        sigchunk, tmp = self.sigma_numba(data,1,rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)
        del tmp
        tmp = (sigma[rowstartread:rowend,:] < sigchunk).choose(sigchunk,sigma[rowstartread:rowend,:])
        sigma[rowstartread:rowend,:] = tmp
      else:
        sigchunk = sigma[rowstartread:rowend,:]

      #dataout = data

      dataout = self.nlpar_nb(data,lam, sr, dthresh, sigchunk,
                              rowcountread,ncols,indices,saturation_protect)

      dataout = dataout.reshape(rowcountread, ncols, phw)
      dataout = dataout[j-rowstartread:, :, : ]
      shpout = dataout.shape
      dataout = dataout.reshape(shpout[0]*shpout[1], pheight, pwidth)
      if rescale == True:
        for i in range(dataout.shape[0]):
          temp = dataout[i,:,:]
          temp -= temp.min()
          temp *= np.float32(mxval)/temp.max()
          dataout[i,:,:] = temp

      patternfileout.write_data(newpatterns=dataout,patStartCount = [[0,j], [ncols, shpout[0]]],
                                     flt2int='clip',scalevalue=1.0 )
      #self.patternfileout.write_data(newpatterns=dataout,patStartCount=[j*ncols,shpout[0]*shpout[1]],
      #                               flt2int='clip',scalevalue=1.0 )
      #return dataout
      #sigma[j:j+rowstartcount[1],:] += \
      #  sigchunk[rowstartcount[0]:rowstartcount[0]+rowstartcount[1],:]


    if verbose >= 2:
      print('', end='')

    numba.set_num_threads(nthreadpos)
    return str(patternfileout.filepath)

  def calcsigma(self,chunksize=0,nn=1,saturation_protect=True,automask=True):
    self.sigmann = nn
    patternfile = self.getinfileobj()


    nrows = np.int64(self.nrows)#np.uint64(patternfile.nRows)
    ncols = np.int64(self.ncols)#np.uint64(patternfile.nCols)

    pwidth = np.uint64(patternfile.patternW)
    pheight = np.uint64(patternfile.patternH)
    phw = pheight*pwidth

    if chunksize <= 0:
      chunksize = np.round(2e9/phw/ncols) # keeps the chunk at about 8GB
      chunksize = np.minimum(nrows,chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)


    nn = np.uint64(nn)

    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth), dtype=np.uint8)

    indices = np.asarray( (self.mask.flatten().nonzero())[0], np.uint64)

    sigma = np.zeros((nrows, ncols), dtype=np.float32)
    #d_nn = np.zeros((nrows, ncols, int((2*nn+1)**2)), dtype=np.float32)
    colstartcount = np.asarray([0,ncols],dtype=np.int64)

    for j in range(0,nrows,chunksize):
      rowstartread = np.int64(max(0, j-nn))
      rowend = min(j + chunksize+nn,nrows)
      if (rowend - rowstartread) < (3):
        rowstartread = np.int64(max(0, rowend - (3)))
      rowcountread = np.int64(rowend-rowstartread)
      data, xyloc = patternfile.read_data(patStartCount = [[0,rowstartread], [ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shp = data.shape
      data = data.reshape(data.shape[0], phw)

      #data = None
      if rowend == nrows:
        rowstartcount = np.asarray([j-rowstartread,rowcountread - (j-rowstartread) ], dtype=np.int64)
      else:
        rowstartcount = np.asarray([j-rowstartread,chunksize ], dtype=np.int64)

      sigchunk, temp = self.sigma_numba(data,nn, rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)

      sigma[j:j+rowstartcount[1],:] += \
        sigchunk[rowstartcount[0]:rowstartcount[0]+rowstartcount[1],:]

    return sigma

  def backsub(self, data):
    # This function will fit a 2D gaussian on top of a plane to the averaged set of patterns (data) that is provided.
    # It will automatically use whatever mask is defined for valid data.
    # If the gaussian fit fails to converge, it will fall back to just using the mean set of patterns for the background
    # with a warning.


    def gaussian_surf(x, y, a, x0, y0, sigx, sigy, c, d, e):
    # equation for 2D gaussian on top of a plane.
      return a * np.exp(- ((x - x0) ** 2) / (2.0 * sigx ** 2) - ((y - y0) ** 2) / (2.0 * sigy ** 2)) + c + d * x + e * y

    def fit_gauss(M, *args):
    # helper function
      x, y = M
      #arr = np.zeros(x.shape)
      return gaussian_surf(x, y, *args)

    back = np.mean(data, axis=0) # start with the mean of all the data
    # now fit a 2D gaussian sitting on a plane.  See fuction def above.
    nx = data.shape[-1]
    ny = data.shape[-2]
    x = np.arange(nx, dtype=float)
    x = (np.broadcast_to(x.reshape(1,nx), (ny, nx)))
    y = np.arange(ny, dtype=float)
    y = (np.broadcast_to(y, (nx, ny)).T)
    x = x.ravel()
    y = y.ravel()

    # need to grab only the values that are in the mask.
    wh = np.nonzero(self.mask.ravel())[0]
    xwh = x[wh]
    ywh = y[wh]
    xywh = np.vstack((xwh, ywh))
    zwh = (back.ravel())[wh]
    whmx = np.unravel_index(back.argmax(), back.shape)
    minz = zwh.min()
    # initialize a guess for the parameters.
    # [gauss amplitude, max loc x, max loc y, sigx, sigy, const offset, slope x, slope y]
    p0 = [(zwh.max() - zwh.min()), whmx[1], whmx[0], nx/2.355, ny/2.355, minz, 0, 0]
    try:
      popt, pcov = opt.curve_fit(fit_gauss, xywh, zwh, p0)
      backfit = (gaussian_surf(x, y, *popt)).reshape(ny, nx)
      #print(p0, popt)
    except RuntimeError:
      print('Warning: no convergence on back subtract ... using mean of the patterns.')
      print('This may not be ideal for scans with few grains across the width of the scan.')
      backfit = back
    backfit -= np.mean(backfit)
    #f, axarr = plt.subplots(1, 3)
    #f.set_size_inches(10, 4)
    #axarr[0].imshow(data[0,:,:].squeeze(), cmap='gray')
    #axarr[1].imshow(data[0,:,:].squeeze() - backfit, cmap='gray')
    #axarr[2].imshow(backfit, cmap='gray')

    data -= backfit
    return data

  @staticmethod
  def automask( h,w ):
    r = (min(h,w)*0.98*0.5)
    x = np.arange(w, dtype=np.float32)
    x = np.minimum(x , (w-x))
    x = x.reshape(1,w)
    y = np.arange(h, dtype=np.float32)
    y = np.minimum(y , (h - y))
    y = y.reshape(h,1)

    mask =np.sqrt(y**2 + x**2)
    mask = (mask < r).astype(np.uint8)
    mask = np.roll(mask, (int(h/2), int(w/2) ), axis=(0,1))

    return mask

  @staticmethod
  @numba.jit(nopython=True,cache=True,fastmath=False,parallel=True)
  def sigma_numba(data, nn, nrows, ncols, rowstartcount, colstartcount, indices, saturation_protect=True):
    sigma = np.zeros((nrows,ncols), dtype = np.float32)
    dout = np.zeros((nrows,ncols,(nn*2+1)**2), dtype=np.float32)
    nout = np.zeros((nrows,ncols,(nn * 2 + 1) ** 2),dtype=np.float32)
    dij = np.zeros((nrows,ncols,(nn * 2 + 1) ** 2, 2),dtype=np.uint64)
    shpdata = data.shape

    n0 = np.float32(shpdata[-1])
    shpind = indices.shape
    mxval = np.max(data)
    if saturation_protect == False:
      mxval += 1.0
    else:
      mxval *= 0.9961

    for j in numba.prange(rowstartcount[0], rowstartcount[0]+rowstartcount[1]):
      nn_r_start = j - nn if (j - nn) >= 0 else 0
      nn_r_end = (j + nn if (j + nn) < nrows else nrows-1)+1
      for i in numba.prange(colstartcount[0], colstartcount[0]+colstartcount[1]):
        nn_c_start = i - nn if (i - nn) >= 0 else 0
        nn_c_end = (i + nn if (i + nn) < ncols else ncols - 1) + 1

        mind = np.float32(1e24)
        indx_0 = i+ncols*j
        count = 0
        for j_nn in range(nn_r_start,nn_r_end ):
          for i_nn in range(nn_c_start,nn_c_end):
            dij[j,i,count,0] = np.uint64(j_nn)
            dij[j,i,count,1] = np.uint64(i_nn) # want to save this for labmda optimization
            indx_nn = i_nn+ncols*j_nn
            d2 = np.float32(0.0)
            n2 = np.float32(1.0e-12)
            nout[j,i,count] = n0 # want to save this for labmda optimization
            if not((i == i_nn) and (j == j_nn)):
              for q in range(shpind[0]):
                d0 = data[indx_0, indices[q]]
                d1 = data[indx_nn, indices[q]]
                if (d1 < mxval) and (d0 < mxval):
                  n2 += 1.0
                  d2 += (d0 - d1)**2
              nout[j,i,count] = n2

              if d2 >= 1.e-3: #sometimes EDAX collects the same pattern twice
                s0 = d2 / np.float32(n2 * 2.0)
                if s0 < mind:
                  mind = s0
            dout[j,i,count] = d2 # want to save this for labmda optimization

            count += 1

        sigma[j,i] = np.sqrt(mind)
        #if sigma[j,i] > 1e12:
        #  print(sigma[j,i], dout[j,i,:], nout[i,j,:])
    return sigma,( dout, nout, dij)

  @staticmethod
  @numba.jit(nopython=True,cache=True,fastmath=False,parallel=True)
  def nlpar_nb(data,lam, sr, dthresh, sigma, nrows,ncols,indices,saturation_protect=True):
    def getpairid(idx0, idx1):
      idx0_t = int(idx0)
      idx1_t = int(idx1)
      if idx0 < idx1:
        pairid = idx0_t + (idx1_t << 32)
      else:
        pairid = idx1_t + (idx0_t << 32)
      return numba.uint64(pairid)

    lam2 = 1.0 / lam**2
    dataout = np.zeros_like(data, np.float32)
    shpdata = data.shape
    shpind = indices.shape
    winsz = np.int32((2*sr+1)**2)


    mxval = np.max(data)
    if saturation_protect == False:
      mxval += np.float32(1.0)
    else:
      mxval *= np.float32(0.999)
    for i in numba.prange(ncols):
      winstart_x = max((i - sr),0) - max((i + sr - (ncols - 1)),0)
      winend_x = min((i + sr),(ncols - 1)) + max((sr - i),0) + 1
      pairdict = numba.typed.Dict.empty(key_type=numba.core.types.uint64,value_type=numba.core.types.float32)
      for j in range(nrows):
        winstart_y = max((j - sr),0) - max((j + sr - (nrows - 1)),0)
        winend_y = min((j + sr),(nrows - 1)) + max((sr - j),0) + 1

        weights = np.zeros(winsz,dtype=np.float32)
        pindx = np.zeros(winsz,dtype=np.uint64)

        indx_0 = i + ncols * j
        sigma0 = sigma[j,i]**2

        counter = 0
        for j_nn in range(winstart_y,winend_y):
          for i_nn in range(winstart_x,winend_x):
            indx_nn = i_nn + ncols * j_nn
            pindx[counter] = indx_nn

            if indx_nn == indx_0:
              weights[counter] = np.float32(-1.0e6)
            else:
              pairid = getpairid(indx_0, indx_nn)
              if pairid in pairdict:
                weights[counter] = pairdict[pairid]
              else:
                sigma1 = sigma[j_nn,i_nn]**2
                d2 = np.float32(0.0)
                n2 = np.float32(0.0)
                for q in range(shpind[0]):
                  d0 = data[indx_0,indices[q]]
                  d1 = data[indx_nn,indices[q]]
                  if (d1 < mxval) and (d0 < mxval):
                    n2 += np.float32(1.0)
                    d2 += (d0 - d1) ** np.int32(2)
                d2 -= n2*(sigma0+sigma1)
                dnorm = (sigma1 + sigma0)*np.sqrt(np.float32(2.0)*n2)
                if dnorm > np.float32(1.e-8):
                  d2 /= dnorm
                else:
                  d2 = np.float32(1e6)*n2
                weights[counter] = d2
                pairdict[pairid] = numba.float32(d2)
            counter += 1
        #print('________________')
        # end of window scanning
        sum = np.float32(0.0)
        for i_nn in range(winsz):

          weights[i_nn] = np.maximum(weights[i_nn]-dthresh, numba.float32(0.0))
          weights[i_nn] = np.exp(-1.0 * weights[i_nn] * lam2)
          sum += weights[i_nn]

        for i_nn in range(winsz):
          indx_nn = pindx[i_nn]
          weights[i_nn] /= sum
          #print(weights[i_nn], ' \n')
          for q in range(shpdata[1]):
            dataout[indx_0, q] += data[indx_nn, q]*weights[i_nn]
        #print('_______', '\n')
    return dataout

  def _calcchunks(self, patdim, ncol, nrow, target_bytes=2e9, col_overlap=0, row_overlap=0, col_offset=0, row_offset=0):

    col_overlap = min(col_overlap, ncol - 1)
    row_overlap = min(row_overlap, nrow - 1)

    byteperpat = patdim[-1] * patdim[-2] * 4 * 2  # assume a 4 byte float input and output array
    byteperdataset = byteperpat * ncol * nrow
    nchunks = int(np.ceil(byteperdataset / target_bytes))

    ncolchunks = (max(np.round(np.sqrt(nchunks * float(ncol) / nrow)), 1))
    colstep = max((ncol / ncolchunks), 1)
    ncolchunks = max(ncol / colstep, 1)
    colstepov = min(colstep + 2 * col_overlap, ncol)
    ncolchunks = max(int(np.ceil(ncolchunks)), 1)
    colstep = max(int(np.round(colstep)), 1)
    colstepov = min(colstep + 2 * col_overlap, ncol)

    nrowchunks = max(np.ceil(nchunks / ncolchunks), 1)
    rowstep = max((nrow / nrowchunks), 1)
    nrowchunks = max(nrow / rowstep, 1)
    rowstepov = min(rowstep + 2 * row_overlap, nrow)
    nrowchunks = max(int(np.ceil(nrowchunks)), 1)
    rowstep = max(int(np.round(rowstep)), 1)
    rowstepov = min(rowstep + 2 * row_overlap, nrow)

    # colchunks = np.round(np.arange(ncolchunks+1)*ncol/ncolchunks).astype(int)
    # colchunks = np.zeros((ncolchunks, 2), dtype=int)
    # colchunks[:, 0] = (np.arange(ncolchunks) * colstep).astype(int)
    # colchunks[:, 1] = colchunks[:, 0] + colstepov - int(col_overlap)
    # colchunks[:, 0] -= col_overlap
    # colchunks[0, 0] = 0;

    colchunks = []
    col_overlap = int(col_overlap)
    for c in range(ncolchunks):
      cchunk = [int(c * colstep) - col_overlap, int(c * colstep + colstepov) - col_overlap]
      colchunks.append(cchunk)
      if cchunk[1] > ncol:
        break

    ncolchunks = len(colchunks)
    colchunks = np.array(colchunks, dtype=int)
    colchunks[0, 0] = 0
    colchunks[-1, 1] = ncol

    if ncolchunks > 1:
      colchunks[-1, 0] = max(0, colchunks[-2, 1] -  2*col_overlap-1)

    colchunks += col_offset

    # for i in range(ncolchunks - 1):
    #   if colchunks[i + 1, 0] >= ncol:
    #     colchunks = colchunks[0:i + 1, :]

    rowchunks = []
    row_overlap = int(row_overlap)
    for r in range(nrowchunks):
      rchunk = [int(r * rowstep) - row_overlap, int(r * rowstep + rowstepov) - row_overlap]
      rowchunks.append(rchunk)
      if rchunk[1] > nrow:
        break

    nrowchunks = len(rowchunks)
    rowchunks = np.array(rowchunks, dtype=int)
    rowchunks[0, 0] = 0
    # for i in range(ncolchunks - 1):
    #   if colchunks[i + 1, 0] >= ncol:
    #     colchunks = colchunks[0:i + 1, :]

    rowchunks = []
    row_overlap = int(row_overlap)
    for r in range(nrowchunks):
      rchunk = [int(r * rowstep) - row_overlap, int(r * rowstep + rowstepov) - row_overlap]
      rowchunks.append(rchunk)
      if rchunk[1] > nrow:
        break

    nrowchunks = len(rowchunks)
    rowchunks = np.array(rowchunks, dtype=int)
    rowchunks[0, 0] = 0
    rowchunks[-1, 1] = nrow

    if nrowchunks > 1:
      rowchunks[-1, 0] = max(0, rowchunks[-2, 1] -  2*row_overlap-1)

    rowchunks += row_offset

    return ncolchunks, nrowchunks, colchunks, rowchunks

  # def asciiupdate(self, nrow, ncol, completematrix):
  #   cm = completematrix
  #   ncdisplay = min(ncol, 80)
  #   nrow





# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

from itertools import product

import numpy as np

from pyebsdindex import tripletvote


class TestAddPhase:
    def test_add_phase_fcc(self):
        phase = tripletvote.addphase("FCC")
        assert np.allclose(
            phase.polefamilies, [[0, 0, 2], [1, 1, 1], [0, 2, 2], [1, 1, 3]]
        )
        angles = phase.angpairs["angles"]
        assert angles.size == 21
        assert np.unique(angles).size == 17

    def test_add_phase_bcc(self):
        phase = tripletvote.addphase("BCC")
        assert np.allclose(
            phase.polefamilies, [[0, 1, 1], [0, 0, 2], [1, 1, 2], [0, 1, 3]]
        )
        angles = phase.angpairs["angles"]
        assert angles.size == 34
        assert np.unique(angles).size == 28

    def test_add_phase_hcp(self):
        phase = tripletvote.addphase("HCP")
        assert np.allclose(
            phase.polefamilies,
            [
                [1, 0, -1, 0],
                [0, 0,  0, 2],
                [1, 0, -1, 1],
                [1, 0, -1, 2],
                [1, 1, -2, 0],
                [1, 0, -1, 3],
                [1, 1, -2, 2],
                [2, 0, -2, 1],
            ]
        )
        angles = phase.angpairs["angles"]
        assert angles.size == 82
        assert np.unique(angles).size == 74

    def test_add_phase_triclinic(self):
        # Build our own reflector list
        hkl = [1, 1, 1]
        hkl_ranges = [np.arange(-i, i + 1) for i in hkl]
        hkl = np.asarray(list(product(*hkl_ranges)), dtype=int)
        hkl = hkl[~np.all(hkl == 0, axis=1)]  # Remove (000)

        phase = tripletvote.addphase(
            spacegroup=1,
            latticeparameter=[2, 3, 4, 70, 100, 120],
            nband_earlyexit=5,
            polefamilies=hkl,
        )

        assert phase.spacegroup == 1
        assert np.allclose(phase.latticeparameter, [2, 3, 4, 70, 100, 120])
        assert phase.nband_earlyexit == 5
        assert phase.npolefamilies == hkl.shape[0] // 2
        for hkl_i in phase.polefamilies:
            assert hkl_i in hkl

        angles = phase.angpairs["angles"]
        assert angles.size == 78
        assert np.unique(angles).size == 77



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

"""Optimization of the pattern center (PC) of EBSD patterns."""

import numpy as np
import multiprocessing
import functools
import scipy.optimize as opt
from timeit import default_timer as timer


__all__ = [
    "optimize",
    "optimize_pso",
]

RADEG = 180.0 / np.pi


def _optfunction(PC_i, indexer, banddat):
    tic = timer()
    PC = np.atleast_2d(PC_i)
    result = np.zeros(PC.shape[0])
    # this loop is here because pyswarms expects a vectorized function
    #print(PC.shape)
    for q in range(PC.shape[0]):

        bandnorm = indexer.bandDetectPlan.radonPlan.radon2pole(
            banddat, PC=PC[q,:], vendor=indexer.vendor
        )
        #print(timer() - tic)
        npoints = banddat.shape[0]
        #n_averages = 0
        #average_fit = 0
        #nbands_fit = 0
        #phase = indexer.phaseLib[0]
        nbands = indexer.bandDetectPlan.nBands
        indexdata, banddat = indexer._indexbandsphase( banddat, bandnorm)



        fit = indexdata[-1]['fit']
        nmatch = indexdata[-1]['nmatch']
        average_fit = fit + 1.0*(nbands - nmatch)
        #average_fit = -1.0*(3.0-fit)*nmatch
        whgood = np.nonzero(fit < 90.0)

        n_averages = len(whgood[0])


        if n_averages < 0.9:
            average_fit = 1000
        else:
            average_fit = np.sum(average_fit[whgood[0]]) + 4.0*(nbands+1)*(npoints - n_averages)
            average_fit /= npoints
            #average_fit /= n_averages
            #average_fit *=  (n_averages*(nbands+1) - nbands_fit)/(n_averages*nbands)
        result[q] = average_fit
    #print(timer()-tic)
    return result


def optimize(pats, indexer, PC0=None, batch=False):
    """Optimize pattern center (PC) (PCx, PCy, PCz) in the convention
    of the :attr:`indexer.vendor` with Nelder-Mead.

    Parameters
    ----------
    pats : numpy.ndarray
        EBSD pattern(s), of shape
        ``(n detector rows, n detector columns)``,
        or ``(n patterns, n detector rows, n detector columns)``.
    indexer : pyebsdindex.ebsd_index.EBSDIndexer
        EBSD indexer instance storing all relevant parameters for band
        detection.
    PC0 : list, optional
        Initial guess of PC. If not given, :attr:`indexer.PC` is used.
        If :attr:`indexer.vendor` is ``"EMSOFT"``, the PC must be four
        numbers, the final number being the pixel size.
    batch : bool, optional
        Default is ``False`` which indicates the fit for a set of
        patterns should be optimized using the cumulative fit for all
        the patterns, and one PC will be returned. If ``True``, then an
        optimization is run for each individual pattern, and an array of
        PC values is returned.

    Returns
    -------
    numpy.ndarray
        Optimized PC.

    Notes
    -----
    SciPy's Nelder-Mead minimization function is used with a tolerance
    ``fatol=0.00001`` between each iteration, ending the optimization
    when the improvement is below this value.
    """
    banddat = indexer.bandDetectPlan.find_bands(pats)
    npoints, nbands = banddat.shape[:2]

    if PC0 is None:
        PC0 = indexer.PC
    emsoftflag = False
    if indexer.vendor == "EMSOFT":  # Convert to EDAX for optimization
        emsoftflag = True
        indexer.vendor = "EDAX"
        patDim = np.array(indexer.bandDetectPlan.patDim)
        delta = indexer.PC
        PCtemp = PC0[0:3]
        patdimnorm = (np.array([patDim[1], patDim[0], np.max(patDim[0:2])]))
        PCtemp[0] *= -1.0
        PCtemp[0] += 0.5 * indexer.bandDetectPlan.patDim[1]
        PCtemp[1] += 0.5 * indexer.bandDetectPlan.patDim[0]
        #PCtemp /= indexer.bandDetectPlan.patDim[1]
        PCtemp /= patdimnorm
        PCtemp[2] /= delta[3]
        PC0 = PCtemp

    if not batch:
        PCopt = opt.minimize(
            _optfunction,
            PC0,
            args=(indexer, banddat),
            method="Nelder-Mead",
            options={"fatol": 0.00001}
        )
        PCoutRet = PCopt['x']
    else:
        PCoutRet = np.zeros((npoints, 3))
        for i in range(npoints):
            PCopt = opt.minimize(
                _optfunction,
                PC0,
                args=(indexer, banddat[i, :].reshape(1, nbands)),
                method="Nelder-Mead"
            )
            PCoutRet[i, :] = PCopt['x']

    if emsoftflag:  # Return original state for indexer
        indexer.vendor = "EMSOFT"
        indexer.PC = delta
        patDim = np.array(indexer.bandDetectPlan.patDim)
        patdimnorm = (np.array([patDim[1], patDim[0], np.max(patDim[0:2])]))
        if PCoutRet.ndim == 2:
            newout = np.zeros((npoints, 4))
            PCoutRet[:, 0] -= 0.5
            #PCoutRet[:, :3] *= indexer.bandDetectPlan.patDim[1]
            PCoutRet[:, :3] *= np.atleast_2d(patdimnorm)
            PCoutRet[:, 0] *= -1.0
            PCoutRet[:, 1] -= 0.5 * patDim[0]
            PCoutRet[:, 2] *= delta[3]
            newout[:, :3] = PCoutRet
            newout[:, 3] = delta[3]
            PCoutRet = newout
        else:
            newout = np.zeros(4)
            PCoutRet[0] -= 0.5
            PCoutRet[:3] *= patdimnorm
            #PCoutRet[:3] *= indexer.bandDetectPlan.patDim[1]
            PCoutRet[1] -= 0.5 * patDim[0]
            PCoutRet[0] *= -1.0
            PCoutRet[2] *= delta[3]
            newout[:3] = PCoutRet
            newout[3] = delta[3]
            PCoutRet = newout

    return PCoutRet


def optimize_pso(
    pats,
    indexer,
    PC0=None,
    batch=False,
    search_limit=0.2,
    early_exit = 0.0001,
    nswarmparticles=30,
    pswarmpar=None,
    niter=50,
    return_cost=False,
    verbose=1
):
    """Optimize pattern center (PC) (PCx, PCy, PCz) in the convention
    of the :attr:`indexer.vendor` with particle swarms.

    Parameters
    ----------
    pats : numpy.ndarray
        EBSD pattern(s), of shape
        ``(n detector rows, n detector columns)``,
        or ``(n patterns, n detector rows, n detector columns)``.
    indexer : pyebsdindex.ebsd_index.EBSDIndexer
        EBSD indexer instance storing all relevant parameters for band
        detection.
    PC0 : list, optional
        Initial guess of PC. If not given, :attr:`indexer.PC` is used.
        If :attr:`indexer.vendor` is ``"EMSOFT"``, the PC must be four
        numbers, the final number being the pixel size.
    batch : bool, optional
        Default is ``False`` which indicates the fit for a set of
        patterns should be optimized using the cumulative fit for all
        the patterns, and one PC will be returned. If ``True``, then an
        optimization is run for each individual pattern, and an array of
        PC values is returned.
    search_limit : float, optional
        Default is 0.2 for all PC values, and sets the +/- limit for the
        optimization search.
    early_exit: float, optional
        Default is 0.0001 for all PC values, and sets a value for which
        the optimum is considered converged before the number of iterations
        is reached.  The optimiztion will exit early if the velocity and distance
        of all the swarm particles is less than the early_exit value.
    nswarmparticles : int, optional
        Number of particles in a swarm. Default is 30.
    pswarmpar : dict, optional
        Particle swarm parameters "c1", "c2", and "w" with defaults 3.5,
        3.5, and 0.8, respectively.
    niter : int, optional
        Number of iterations. Default is 50.
    return_costs: bool, optional
        Set to True to return the cost value as well as the optimum fit PC.
    verbose : int, optional
        Whether to print the parameters and progress of the
        optimization (>= 1) or not (< 1). Default is to print.

    Returns
    -------
    numpy.ndarray
        Optimized PC.
    """
    banddat = indexer.bandDetectPlan.find_bands(pats)
    npoints, nbands = banddat.shape[:2]
    if pswarmpar is None:
        #pswarmpar = {"c1": 3.05, "c2": 1.05, "w": 0.8}
        pswarmpar = {"c1": 3.5, "c2": 3.5, "w": 0.8}
    if nswarmparticles is None:
        #nswarmpoints = int(np.array(search_limit).max() * (10.0/0.2))
        nswarmparticles = 30

    nswarmparticles = max(5, nswarmparticles)

    if PC0 is None:
        PC0 = np.asarray(indexer.PC)
    else:
        PC0 = np.asarray(PC0)

    emsoftflag = False
    if indexer.vendor == "EMSOFT":  # Convert to EDAX for optimization
        emsoftflag = True
        indexer.vendor = "EDAX"
        delta = indexer.PC
        PCtemp = PC0[0:3]
        PCtemp[0] *= -1.0
        PCtemp[0] += 0.5 * indexer.bandDetectPlan.patDim[1]
        PCtemp[1] += 0.5 * indexer.bandDetectPlan.patDim[0]
        PCtemp /= indexer.bandDetectPlan.patDim[1]
        PCtemp[2] /= delta[3]
        PC0 = np.array(PCtemp)



    # optimizer = pso.single.GlobalBestPSO(
    #     n_particles=nswarmpoints,
    #     dimensions=3,
    #     options=pswarmpar,
    #     bounds=(PC0 - np.array(search_limit), PC0 + np.array(search_limit)),
    # )
    optimizer = PSOOpt(dimensions=3, n_particles=nswarmparticles,
                       c1=pswarmpar['c1'],
                       c2 = pswarmpar['c2'], w = pswarmpar['w'], hyperparammethod='auto',
                       early_exit=early_exit)

    if not batch:
        # cost, PCoutRet = optimizer.optimize(
        #     _optfunction, niter, indexer=indexer, banddat=banddat
        # )
        cost, PCoutRet = optimizer.optimize(_optfunction, indexer=indexer, banddat=banddat,
                                            start=PC0, bounds=(PC0 - np.array(search_limit), PC0 + np.array(search_limit)),
                                            niter=niter, verbose=verbose)
        costout = cost
        #print(cost)
    else:
        PCoutRet = np.zeros((npoints, 3))
        if verbose >= 1:
            print('', end='\n')
        costout = np.zeros(npoints, dtype=np.float32)
        for i in range(npoints):
            # cost, PCoutRet[i, :] = optimizer.optimize(
            #     _optfunction, niter, indexer=indexer, banddat=banddat[i, :, :]
            # )


            cost, newPC = optimizer.optimize(_optfunction, indexer=indexer,
                                    banddat=banddat[i, :].reshape(1, nbands),
               start=PC0, bounds=(PC0 - np.array(search_limit), PC0 + np.array(search_limit)),
               niter=niter, verbose=0)

            PCoutRet[i, :] = newPC
            costout[i] = cost
            progress = int(round(10 * float(i) / npoints))
            if verbose >= 1:
                print('', end='\r')
                print('PC found: [',
                      '*' * progress, ' ' * (10 - progress), '] ', i + 1, '/', npoints,
                      '  global best:', "{0:.3g}".format(cost),
                      '  PC opt:', np.array_str(PCoutRet[i,:], precision=4, suppress_small=True),
                      sep='', end='')
        if verbose >= 1:
            print('', end='\n')

    if emsoftflag:  # Return original state for indexer
        indexer.vendor = "EMSOFT"
        indexer.PC = delta
        if PCoutRet.ndim == 2:
            newout = np.zeros((npoints, 4))
            PCoutRet[:, 0] -= 0.5
            PCoutRet[:, :3] *= indexer.bandDetectPlan.patDim[1]
            PCoutRet[:, 1] -= 0.5 * indexer.bandDetectPlan.patDim[0]
            PCoutRet[:, 0] *= -1.0
            PCoutRet[:, 2] *= delta[3]
            newout[:, :3] = PCoutRet
            newout[:, 3] = delta[3]
            PCoutRet = newout
        else:
            newout = np.zeros(4)
            PCoutRet[0] -= 0.5
            PCoutRet[:3] *= indexer.bandDetectPlan.patDim[1]
            PCoutRet[1] -= 0.5 * indexer.bandDetectPlan.patDim[0]
            PCoutRet[0] *= -1.0
            PCoutRet[2] *= delta[3]
            newout[:3] = PCoutRet
            newout[3] = delta[3]
            PCoutRet = newout
    if return_cost is False:
        return PCoutRet
    else:
        return PCoutRet, costout

def _file_opt(fobj, indexer, stride=200, groupsz = 3):
    nCols = fobj.nCols
    nRows = fobj.nRows
    pcopt = np.zeros((int(nRows / stride), int(nCols / stride), 3), dtype=np.float32)

    for i in range(int(nRows / stride)):
        ii = i * stride
        print(ii)
        for j in range(int(nCols / stride)):
            jj = j * stride

            pats = fobj.read_data(
                returnArrayOnly=True,
                convertToFloat=True,
                patStartCount=[ii*nCols + jj, groupsz]
            )

            pc = optimize(pats, indexer)
            pcopt[i, j, :] = pc

    return pcopt


class PSOOpt():
    def __init__(self,
                dimensions=3,
                n_particles=50,
                c1 = 2.05,
                c2 = 2.05,
                w = 0.8,
                hyperparammethod = 'static',
                boundmethod = 'bounce',
                early_exit=None):
        self.n_particles = int(n_particles)
        self.dimensions = int(dimensions)
        self.c1 = c1
        self.c2 = c2
        self.c1i = None
        self.c2i = None
        self.w = w
        self.wi = None
        self.hyperparammethod = hyperparammethod
        self.boundmethod = boundmethod
        self.vellimit = None
        self.start = None
        self.bounds = None
        self.range = None
        self.niter = None
        self.pos = None
        self.vel = None
        self.early_exit = early_exit


    def initializeswarm(self, start=None, bounds=None):


        if start is None:
            if bounds is not None:
                start = 0.5*(bounds[0]+bounds[1])
            else:
                start = np.zeros(self.dimensions, dtype=np.float32)

        self.start = start

        if bounds is None:
            bounds = (-1*np.ones(self.dimensions, dtype=np.float32),np.ones(self.dimensions, dtype=np.float32) )

        self.bounds = bounds
        self.range = self.bounds[1] - self.bounds[0]

        self.pos = np.random.uniform(low=bounds[0], high=bounds[1], size=(self.n_particles, self.dimensions))
        self.pos[0, :] = start

        self.vel = np.random.normal(size=(self.n_particles, self.dimensions), loc=0.0, scale=1.0)
        meanv = np.mean(np.sqrt(np.sum(self.vel**2, axis=1)))
        self.vel *= np.sqrt(np.sum(self.range**2))/(20. * meanv)
        self.vellimit = 4*np.mean(np.sqrt(np.sum(self.vel**2, axis=1)))


        self.pbest = np.zeros(self.n_particles) + np.infty
        self.pbest_loc = np.copy(self.pos)
        self.gbest = np.infty
        self.gbest_loc = start




    def updateswarmbest(self, fun2opt, pool, **kwargs):

        val = np.zeros(self.n_particles)

        #tic = timer()
        for part_i in range(self.n_particles):
            temp = self.pos[part_i, :]
            val[part_i] = fun2opt(temp, **kwargs)
        #print(timer()-tic)
        #pos = self.pos.copy()
        #tic = timer()
        #results = pool.map(functools.partial(fun2opt, **kwargs),list(pos) )
        #print(timer()-tic)
        #print(len(results[0]), type(results[0]))
        #print(len(results))
        #val = np.concatenate(results)

        wh_newpbest = np.nonzero(val < self.pbest)[0]

        self.pbest[wh_newpbest] = val[wh_newpbest]
        self.pbest_loc[wh_newpbest, :] = self.pos[wh_newpbest, :]

        wh_minpbest = np.argmin(self.pbest)
        if self.pbest[wh_minpbest] < self.gbest:
            self.gbest = self.pbest[wh_minpbest]
            self.gbest_loc = self.pbest_loc[wh_minpbest, :]


    def updateswarmvelpos(self):

        w = self.wi
        c1 = self.c1i
        c2 = self.c2i
        r1 = np.random.random((self.n_particles,1))
        r2 = np.random.random((self.n_particles,1))
        nvel = self.vel.copy()
        nvel = w * nvel + \
               c1 * r1 * (self.pbest_loc - self.pos) + \
               c2 * r2 * (self.gbest_loc - self.pos)

        mag = np.expand_dims(np.sqrt(np.sum(nvel**2, axis=1)), axis=1)
        wh_toofast = np.nonzero(mag > self.vellimit)[0]
        #print(nvel.shape, wh_toofast.shape, mag.shape)
        nvel[wh_toofast, :] *= self.vellimit/mag[wh_toofast]

        self.vel = nvel
        self.pos += nvel

        self.boundarycheck()





    def boundarycheck(self):

        if str.lower(self.boundmethod) == 'bounce':
            self.boundarybounce()


    def boundarybounce(self):

        lb,ub = self.bounds
        for d in range(self.dimensions):
            wh_under = np.nonzero(self.pos[:,d] < lb[d])[0]
            self.pos[wh_under,d] = lb[d]
            self.vel[wh_under,d] = np.abs(self.vel[wh_under,d])

            wh_over = np.nonzero(self.pos[:, d] > ub[d])[0]
            self.pos[wh_over, d] = ub[d]
            self.vel[wh_over, d] = -1*np.abs(self.vel[wh_over, d])

    def updatehyperparam(self, iter):
        if str.lower(self.hyperparammethod) == 'auto':

            N = float(self.niter)-1
            self.c1i = (self.c1 - self.c1 / 7) * (N-iter)/N + self.c1 / 7.0
            #self.c2i = (self.c1 - self.c1 / 7) * (iter) / N + self.c1 / 7.0
            self.c2i = (self.c2 - self.c2 / 7) * (iter) / N + self.c2 / 7.0
            self.wi = self.w/2 * ((N - iter)/N)**2 + self.w/2

        else:
            self.c1i = self.c1
            self.c2i = self.c2
            self.wi = self.w


        pass
    def printprogress(self, iter):

        progress = int(round(10*float(iter)/self.niter))
        print('',end='\r' )
        print('Progress [',
              '*' * progress, ' '*(10-progress),'] ', iter+1 , '/', self.niter,
              '  global best:', "{0:.3g}".format(self.gbest),
              '  best loc:', np.array_str(self.gbest_loc, precision=4, suppress_small=True),
              sep='', end='')
    def optimize(self, function, start=None, bounds=None, niter=50, verbose = 1, **kwargs):

        self.initializeswarm(start, bounds)
        early_exit = self.early_exit
        if early_exit is None:
           early_exit = -1.0

        #with multiprocessing.get_context("spawn").Pool(min(multiprocessing.cpu_count(), self.n_particles)) as pool:
        pool = None
        if verbose >= 1:
            print('n_particles:', self.n_particles, 'c1:', self.c1, 'c2:', self.c2, 'w:', self.w )

        self.niter = niter
        for iter in range(niter):
            self.updatehyperparam(iter)
            self.updateswarmbest(function, pool, **kwargs)
            if verbose >= 1:
                self.printprogress(iter)
                #print(np.abs(self.vel).max())
            self.updateswarmvelpos()

            if np.abs(self.vel).max() < early_exit:
                d = abs(self.gbest_loc - self.pos)
                    #print(d.max())
                if d.max() < early_exit:
                    break




        #pool.close()
        #pool.terminate()
        final_best = self.gbest
        final_loc = self.gbest_loc
        if verbose >= 1:
            print('', end='\n')
            print("Optimization finished | best cost: {}, best pos: {}".format(
                final_best, final_loc))
            print(' ')
        return final_best, final_loc






# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

"""Tests for internal package technicalities and package
distribution.
"""

import pytest

from pyebsdindex import _pyopencl_installed, _ray_installed


@pytest.mark.skipif(not _pyopencl_installed, reason="pyopencl is not installed")
def test_available_functionality_without_pyopencl():
    from pyebsdindex.band_detect import BandDetect
    from pyebsdindex.opencl.band_detect_cl import BandDetect as BandDetectCL
    assert issubclass(BandDetectCL, BandDetect)


@pytest.mark.skipif(_pyopencl_installed, reason="pyopencl is installed")
def test_unavailable_functionality_without_pyopencl():
    with pytest.raises(ImportError):

        from pyebsdindex.opencl import openclparam
        try:
            clparam = openclparam.OpenClParam()
            gpu = clparam.get_gpu()
            if len(gpu) < 1:
                raise ImportError('')
        except:
            raise ImportError('')

@pytest.mark.skipif(not _ray_installed, reason="ray is not installed")
def test_available_functionality_with_ray():
    from pyebsdindex.ebsd_index import index_pats_distributed
    #from pyebsdindex.ebsd_index import IndexerRay

    #assert callable(index_pats_distributed)
    #_ = IndexerRay.remote()


@pytest.mark.skipif(_ray_installed, reason="ray is installed")
def test_unavailable_functionality_without_ray():
    with pytest.raises(ImportError):
        from pyebsdindex.ebsd_index import index_pats_distributed
    #with pytest.raises(ImportError):
    #    from pyebsdindex.ebsd_index import IndexerRay



####################################################################################################
# Copyright (c) 2017-2020, Martin Diehl/Max-Planck-Institut für Eisenforschung GmbH
# Copyright (c) 2013-2014, Marc De Graef/Carnegie Mellon University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
#
#     - Redistributions of source code must retain the above copyright notice, this list
#        of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright notice, this
#        list of conditions and the following disclaimer in the documentation and/or
#        other materials provided with the distribution.
#     - Neither the names of Marc De Graef, Carnegie Mellon University nor the names
#        of its contributors may be used to endorse or promote products derived from
#        this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################################################################################################

# Note: An object oriented approach to use this conversions is available in DAMASK, see
# https://damask.mpie.de and https://github.com/eisenforschung/DAMASK

import numpy as np

P = 1

# parameters for conversion from/to cubochoric
sc   = np.pi**(1./6.)/6.**(1./6.)
beta = np.pi**(5./6.)/6.**(1./6.)/2.
R1   = (3.*np.pi/4.)**(1./3.)
#---------- Quaternion ----------

def qu2om(qu):
    qq = qu[...,0:1]**2-(qu[...,1:2]**2 + qu[...,2:3]**2 + qu[...,3:4]**2)
    om = np.block([qq + 2.0*qu[...,1:2]**2,
                   2.0*(qu[...,2:3]*qu[...,1:2]-P*qu[...,0:1]*qu[...,3:4]),
                   2.0*(qu[...,3:4]*qu[...,1:2]+P*qu[...,0:1]*qu[...,2:3]),
                   2.0*(qu[...,1:2]*qu[...,2:3]+P*qu[...,0:1]*qu[...,3:4]),
                   qq + 2.0*qu[...,2:3]**2,
                   2.0*(qu[...,3:4]*qu[...,2:3]-P*qu[...,0:1]*qu[...,1:2]),
                   2.0*(qu[...,1:2]*qu[...,3:4]-P*qu[...,0:1]*qu[...,2:3]),
                   2.0*(qu[...,2:3]*qu[...,3:4]+P*qu[...,0:1]*qu[...,1:2]),
                   qq + 2.0*qu[...,3:4]**2,
                  ]).reshape(qu.shape[:-1]+(3,3))
    return om


def qu2eu(qu):
    """Quaternion to Bunge-Euler angles."""
    q02   = qu[...,0:1]*qu[...,2:3]
    q13   = qu[...,1:2]*qu[...,3:4]
    q01   = qu[...,0:1]*qu[...,1:2]
    q23   = qu[...,2:3]*qu[...,3:4]
    q03_s = qu[...,0:1]**2+qu[...,3:4]**2
    q12_s = qu[...,1:2]**2+qu[...,2:3]**2
    chi = np.sqrt(q03_s*q12_s)

    eu = np.where(np.abs(q12_s) < 1.0e-8,
            np.block([np.arctan2(-P*2.0*qu[...,0:1]*qu[...,3:4],qu[...,0:1]**2-qu[...,3:4]**2),
                      np.zeros(qu.shape[:-1]+(2,))]),
                  np.where(np.abs(q03_s) < 1.0e-8,
                      np.block([np.arctan2(   2.0*qu[...,1:2]*qu[...,2:3],qu[...,1:2]**2-qu[...,2:3]**2),
                                np.broadcast_to(np.pi,qu.shape[:-1]+(1,)),
                                np.zeros(qu.shape[:-1]+(1,))]),
                      np.block([np.arctan2((-P*q02+q13)*chi, (-P*q01-q23)*chi),
                                np.arctan2( 2.0*chi,          q03_s-q12_s    ),
                                np.arctan2(( P*q02+q13)*chi, (-P*q01+q23)*chi)])
                          )
                 )
    # reduce Euler angles to definition range
    eu[np.abs(eu)<1.e-6] = 0.0
    eu = np.where(eu<0, (eu+2.0*np.pi)%np.array([2.0*np.pi,np.pi,2.0*np.pi]),eu)                # needed?
    return eu


def qu2ax(qu):
    """
    Quaternion to axis angle pair.

    Modified version of the original formulation, should be numerically more stable
    """
    with np.errstate(invalid='ignore',divide='ignore'):
        s = np.sign(qu[...,0:1])/np.sqrt(qu[...,1:2]**2+qu[...,2:3]**2+qu[...,3:4]**2)
        omega = 2.0 * np.arccos(np.clip(qu[...,0:1],-1.0,1.0))
        ax = np.where(np.broadcast_to(qu[...,0:1] < 1.0e-8,qu.shape),
                      np.block([qu[...,1:4],np.broadcast_to(np.pi,qu.shape[:-1]+(1,))]),
                      np.block([qu[...,1:4]*s,omega]))
    ax[np.isclose(qu[...,0],1.,rtol=0.0)] = [0.0, 0.0, 1.0, 0.0]
    return ax


def qu2ro(qu):
    """Quaternion to Rodrigues-Frank vector."""
    with np.errstate(invalid='ignore',divide='ignore'):
        s  = np.linalg.norm(qu[...,1:4],axis=-1,keepdims=True)
        ro = np.where(np.broadcast_to(np.abs(qu[...,0:1]) < 1.0e-12,qu.shape),
                      np.block([qu[...,1:2], qu[...,2:3], qu[...,3:4], np.broadcast_to(np.inf,qu.shape[:-1]+(1,))]),
                      np.block([qu[...,1:2]/s,qu[...,2:3]/s,qu[...,3:4]/s,
                                np.tan(np.arccos(np.clip(qu[...,0:1],-1.0,1.0)))
                               ])
                   )
    ro[np.abs(s).squeeze(-1) < 1.0e-12] = [0.0,0.0,P,0.0]
    return ro


def qu2ho(qu):
    """Quaternion to homochoric vector."""
    with np.errstate(invalid='ignore'):
        omega = 2.0 * np.arccos(np.clip(qu[...,0:1],-1.0,1.0))
        ho = np.where(np.abs(omega) < 1.0e-12,
                      np.zeros(3),
                      qu[...,1:4]/np.linalg.norm(qu[...,1:4],axis=-1,keepdims=True) \
                      * (0.75*(omega - np.sin(omega)))**(1./3.))
    return ho


def qu2cu(qu):
    """Quaternion to cubochoric vector."""
    return ho2cu(qu2ho(qu))


#---------- Rotation matrix ----------

def om2qu(om):
    """
    Rotation matrix to quaternion.

    This formulation is from  www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion.
    The original formulation had issues.
    """
    trace = om[...,0,0:1]+om[...,1,1:2]+om[...,2,2:3]

    with np.errstate(invalid='ignore',divide='ignore'):
        s = [
             0.5 / np.sqrt( 1.0 + trace),
             2.0 * np.sqrt( 1.0 + om[...,0,0:1] - om[...,1,1:2] - om[...,2,2:3]),
             2.0 * np.sqrt( 1.0 + om[...,1,1:2] - om[...,2,2:3] - om[...,0,0:1]),
             2.0 * np.sqrt( 1.0 + om[...,2,2:3] - om[...,0,0:1] - om[...,1,1:2] )
             ]
        qu= np.where(trace>0,
                     np.block([0.25 / s[0],
                              (om[...,2,1:2] - om[...,1,2:3] ) * s[0],
                              (om[...,0,2:3] - om[...,2,0:1] ) * s[0],
                              (om[...,1,0:1] - om[...,0,1:2] ) * s[0]]),
                     np.where(om[...,0,0:1] > np.maximum(om[...,1,1:2],om[...,2,2:3]),
                              np.block([(om[...,2,1:2] - om[...,1,2:3]) / s[1],
                                        0.25 * s[1],
                                        (om[...,0,1:2] + om[...,1,0:1]) / s[1],
                                        (om[...,0,2:3] + om[...,2,0:1]) / s[1]]),
                              np.where(om[...,1,1:2] > om[...,2,2:3],
                                       np.block([(om[...,0,2:3] - om[...,2,0:1]) / s[2],
                                                 (om[...,0,1:2] + om[...,1,0:1]) / s[2],
                                                 0.25 * s[2],
                                                 (om[...,1,2:3] + om[...,2,1:2]) / s[2]]),
                                       np.block([(om[...,1,0:1] - om[...,0,1:2]) / s[3],
                                                 (om[...,0,2:3] + om[...,2,0:1]) / s[3],
                                                 (om[...,1,2:3] + om[...,2,1:2]) / s[3],
                                                 0.25 * s[3]]),
                                      )
                             )
                    )*np.array([1,P,P,P])
        qu[qu[...,0]<0] *=-1
    return qu


def om2eu(om):
    """Rotation matrix to Bunge-Euler angles."""
    with np.errstate(invalid='ignore',divide='ignore'):
        zeta = 1.0/np.sqrt(1.0-om[...,2,2:3]**2)
        eu = np.where(np.isclose(np.abs(om[...,2,2:3]),1.0,1e-9),
                      np.block([np.arctan2(om[...,0,1:2],om[...,0,0:1]),
                                np.pi*0.5*(1-om[...,2,2:3]),
                                np.zeros(om.shape[:-2]+(1,)),
                               ]),
                      np.block([np.arctan2(om[...,2,0:1]*zeta,-om[...,2,1:2]*zeta),
                                np.arccos( om[...,2,2:3]),
                                np.arctan2(om[...,0,2:3]*zeta,+om[...,1,2:3]*zeta)
                               ])
                      )
    eu[np.abs(eu)<1.e-8] = 0.0
    eu = np.where(eu<0, (eu+2.0*np.pi)%np.array([2.0*np.pi,np.pi,2.0*np.pi]),eu)
    return eu


def om2ax(om):
    """Rotation matrix to axis angle pair."""
    #return qu2ax(om2qu(om)) # HOTFIX
    diag_delta = -P*np.block([om[...,1,2:3]-om[...,2,1:2],
                               om[...,2,0:1]-om[...,0,2:3],
                               om[...,0,1:2]-om[...,1,0:1]
                             ])
    t = 0.5*(om.trace(axis2=-2,axis1=-1) -1.0).reshape(om.shape[:-2]+(1,))
    w,vr = np.linalg.eig(om)
    # mask duplicated real eigenvalues
    w[np.isclose(w[...,0],1.0+0.0j),1:] = 0.
    w[np.isclose(w[...,1],1.0+0.0j),2:] = 0.
    vr = np.swapaxes(vr,-1,-2)
    ax = np.where(np.abs(diag_delta)<1e-12,
                         np.real(vr[np.isclose(w,1.0+0.0j)]).reshape(om.shape[:-2]+(3,)),
                  np.abs(np.real(vr[np.isclose(w,1.0+0.0j)]).reshape(om.shape[:-2]+(3,))) \
                  *np.sign(diag_delta))
    ax = np.block([ax,np.arccos(np.clip(t,-1.0,1.0))])
    ax[np.abs(ax[...,3])<1.e-8] = [ 0.0, 0.0, 1.0, 0.0]
    return ax


def om2ro(om):
    """Rotation matrix to Rodrigues-Frank vector."""
    return eu2ro(om2eu(om))


def om2ho(om):
    """Rotation matrix to homochoric vector."""
    return ax2ho(om2ax(om))


def om2cu(om):
    """Rotation matrix to cubochoric vector."""
    return ho2cu(om2ho(om))


#---------- Bunge-Euler angles ----------

def eu2qu(eu):
    """Bunge-Euler angles to quaternion."""
    ee = 0.5*eu
    cPhi = np.cos(ee[...,1:2])
    sPhi = np.sin(ee[...,1:2])
    qu = np.block([    cPhi*np.cos(ee[...,0:1]+ee[...,2:3]),
                   -P*sPhi*np.cos(ee[...,0:1]-ee[...,2:3]),
                   -P*sPhi*np.sin(ee[...,0:1]-ee[...,2:3]),
                   -P*cPhi*np.sin(ee[...,0:1]+ee[...,2:3])])
    qu[qu[...,0]<0.0]*=-1
    return qu


def eu2om(eu):
    """Bunge-Euler angles to rotation matrix."""
    c = np.cos(eu)
    s = np.sin(eu)
    om = np.block([+c[...,0:1]*c[...,2:3]-s[...,0:1]*s[...,2:3]*c[...,1:2],
                   +s[...,0:1]*c[...,2:3]+c[...,0:1]*s[...,2:3]*c[...,1:2],
                   +s[...,2:3]*s[...,1:2],
                   -c[...,0:1]*s[...,2:3]-s[...,0:1]*c[...,2:3]*c[...,1:2],
                   -s[...,0:1]*s[...,2:3]+c[...,0:1]*c[...,2:3]*c[...,1:2],
                   +c[...,2:3]*s[...,1:2],
                   +s[...,0:1]*s[...,1:2],
                   -c[...,0:1]*s[...,1:2],
                   +c[...,1:2]
                   ]).reshape(eu.shape[:-1]+(3,3))
    om[np.abs(om)<1.e-12] = 0.0
    return om


def eu2ax(eu):
    """Bunge-Euler angles to axis angle pair."""
    t = np.tan(eu[...,1:2]*0.5)
    sigma = 0.5*(eu[...,0:1]+eu[...,2:3])
    delta = 0.5*(eu[...,0:1]-eu[...,2:3])
    tau   = np.linalg.norm(np.block([t,np.sin(sigma)]),axis=-1,keepdims=True)
    alpha = np.where(np.abs(np.cos(sigma))<1.e-12,np.pi,2.0*np.arctan(tau/np.cos(sigma)))
    with np.errstate(invalid='ignore',divide='ignore'):
        ax = np.where(np.broadcast_to(np.abs(alpha)<1.0e-12,eu.shape[:-1]+(4,)),
                      [0.0,0.0,1.0,0.0],
                      np.block([-P/tau*t*np.cos(delta),
                                -P/tau*t*np.sin(delta),
                                -P/tau*  np.sin(sigma),
                                 alpha
                                ]))
    ax[(alpha<0.0).squeeze()] *=-1
    return ax


def eu2ro(eu):
    """Bunge-Euler angles to Rodrigues-Frank vector."""
    ax = eu2ax(eu)
    ro = np.block([ax[...,:3],np.tan(ax[...,3:4]*.5)])
    ro[ax[...,3]>=np.pi,3] = np.inf
    ro[np.abs(ax[...,3])<1.e-16] = [ 0.0, 0.0, P, 0.0 ]
    return ro


def eu2ho(eu):
    """Bunge-Euler angles to homochoric vector."""
    return ax2ho(eu2ax(eu))


def eu2cu(eu):
    """Bunge-Euler angles to cubochoric vector."""
    return ho2cu(eu2ho(eu))


#---------- Axis angle pair ----------

def ax2qu(ax):
    """Axis angle pair to quaternion."""
    c = np.cos(ax[...,3:4]*.5)
    s = np.sin(ax[...,3:4]*.5)
    qu = np.where(np.abs(ax[...,3:4])<1.e-6,[1.0, 0.0, 0.0, 0.0],np.block([c, ax[...,:3]*s]))
    return qu


def ax2om(ax):
    """Axis angle pair to rotation matrix."""
    c = np.cos(ax[...,3:4])
    s = np.sin(ax[...,3:4])
    omc = 1. -c
    om = np.block([c+omc*ax[...,0:1]**2,
                   omc*ax[...,0:1]*ax[...,1:2] + s*ax[...,2:3],
                   omc*ax[...,0:1]*ax[...,2:3] - s*ax[...,1:2],
                   omc*ax[...,0:1]*ax[...,1:2] - s*ax[...,2:3],
                   c+omc*ax[...,1:2]**2,
                   omc*ax[...,1:2]*ax[...,2:3] + s*ax[...,0:1],
                   omc*ax[...,0:1]*ax[...,2:3] + s*ax[...,1:2],
                   omc*ax[...,1:2]*ax[...,2:3] - s*ax[...,0:1],
                   c+omc*ax[...,2:3]**2]).reshape(ax.shape[:-1]+(3,3))
    return om if P < 0.0 else np.swapaxes(om,-1,-2)


def ax2eu(ax):
    """Rotation matrix to Bunge Euler angles."""
    return om2eu(ax2om(ax))


def ax2ro(ax):
    """Axis angle pair to Rodrigues-Frank vector."""
    ro = np.block([ax[...,:3],
                   np.where(np.isclose(ax[...,3:4],np.pi,atol=1.e-15,rtol=.0),
                            np.inf,
                            np.tan(ax[...,3:4]*0.5))
                  ])
    ro[np.abs(ax[...,3])<1.e-6] = [.0,.0,P,.0]
    return ro


def ax2ho(ax):
    """Axis angle pair to homochoric vector."""
    f = (0.75 * ( ax[...,3:4] - np.sin(ax[...,3:4]) ))**(1.0/3.0)
    ho = ax[...,:3] * f
    return ho


def ax2cu(ax):
    """Axis angle pair to cubochoric vector."""
    return ho2cu(ax2ho(ax))


#---------- Rodrigues-Frank vector ----------

def ro2qu(ro):
    """Rodrigues-Frank vector to quaternion."""
    return ax2qu(ro2ax(ro))


def ro2om(ro):
    """Rodgrigues-Frank vector to rotation matrix."""
    return ax2om(ro2ax(ro))


def ro2eu(ro):
    """Rodrigues-Frank vector to Bunge-Euler angles."""
    return om2eu(ro2om(ro))


def ro2ax(ro):
    """Rodrigues-Frank vector to axis angle pair."""
    with np.errstate(invalid='ignore',divide='ignore'):
        ax = np.where(np.isfinite(ro[...,3:4]),
             np.block([ro[...,0:3]*np.linalg.norm(ro[...,0:3],axis=-1,keepdims=True),2.*np.arctan(ro[...,3:4])]),
             np.block([ro[...,0:3],np.broadcast_to(np.pi,ro[...,3:4].shape)]))
    ax[np.abs(ro[...,3]) < 1.e-8]  = np.array([ 0.0, 0.0, 1.0, 0.0 ])
    return ax


def ro2ho(ro):
    """Rodrigues-Frank vector to homochoric vector."""
    f = np.where(np.isfinite(ro[...,3:4]),2.0*np.arctan(ro[...,3:4]) -np.sin(2.0*np.arctan(ro[...,3:4])),np.pi)
    ho = np.where(np.broadcast_to(np.sum(ro[...,0:3]**2.0,axis=-1,keepdims=True) < 1.e-8,ro[...,0:3].shape),
                  np.zeros(3), ro[...,0:3]* (0.75*f)**(1.0/3.0))
    return ho


def ro2cu(ro):
    """Rodrigues-Frank vector to cubochoric vector."""
    return ho2cu(ro2ho(ro))


#---------- Homochoric vector----------

def ho2qu(ho):
    """Homochoric vector to quaternion."""
    return ax2qu(ho2ax(ho))


def ho2om(ho):
    """Homochoric vector to rotation matrix."""
    return ax2om(ho2ax(ho))


def ho2eu(ho):
    """Homochoric vector to Bunge-Euler angles."""
    return ax2eu(ho2ax(ho))


def ho2ax(ho):
    """Homochoric vector to axis angle pair."""
    tfit = np.array([+1.0000000000018852,      -0.5000000002194847,
                     -0.024999992127593126,    -0.003928701544781374,
                     -0.0008152701535450438,   -0.0002009500426119712,
                     -0.00002397986776071756,  -0.00008202868926605841,
                     +0.00012448715042090092,  -0.0001749114214822577,
                     +0.0001703481934140054,   -0.00012062065004116828,
                     +0.000059719705868660826, -0.00001980756723965647,
                     +0.000003953714684212874, -0.00000036555001439719544])
    hmag_squared = np.sum(ho**2.,axis=-1,keepdims=True)
    hm = hmag_squared.copy()
    s = tfit[0] + tfit[1] * hmag_squared
    for i in range(2,16):
        hm *= hmag_squared
        s  += tfit[i] * hm
    with np.errstate(invalid='ignore'):
        ax = np.where(np.broadcast_to(np.abs(hmag_squared)<1.e-8,ho.shape[:-1]+(4,)),
                      [ 0.0, 0.0, 1.0, 0.0 ],
                      np.block([ho/np.sqrt(hmag_squared),2.0*np.arccos(np.clip(s,-1.0,1.0))]))
    return ax


def ho2ro(ho):
    """Axis angle pair to Rodrigues-Frank vector."""
    return ax2ro(ho2ax(ho))


def ho2cu(ho):
    """
    Homochoric vector to cubochoric vector.

    References
    ----------
    D. Roşca et al., Modelling and Simulation in Materials Science and Engineering 22:075013, 2014
    https://doi.org/10.1088/0965-0393/22/7/075013

    """
    rs = np.linalg.norm(ho,axis=-1,keepdims=True)

    xyz3 = np.take_along_axis(ho,_get_pyramid_order(ho,'forward'),-1)

    with np.errstate(invalid='ignore',divide='ignore'):
        # inverse M_3
        xyz2 = xyz3[...,0:2] * np.sqrt( 2.0*rs/(rs+np.abs(xyz3[...,2:3])) )
        qxy = np.sum(xyz2**2,axis=-1,keepdims=True)

        q2 = qxy + np.max(np.abs(xyz2),axis=-1,keepdims=True)**2
        sq2 = np.sqrt(q2)
        q = (beta/np.sqrt(2.0)/R1) * np.sqrt(q2*qxy/(q2-np.max(np.abs(xyz2),axis=-1,keepdims=True)*sq2))
        tt = np.clip((np.min(np.abs(xyz2),axis=-1,keepdims=True)**2\
            +np.max(np.abs(xyz2),axis=-1,keepdims=True)*sq2)/np.sqrt(2.0)/qxy,-1.0,1.0)
        T_inv = np.where(np.abs(xyz2[...,1:2]) <= np.abs(xyz2[...,0:1]),
                            np.block([np.ones_like(tt),np.arccos(tt)/np.pi*12.0]),
                            np.block([np.arccos(tt)/np.pi*12.0,np.ones_like(tt)]))*q
        T_inv[xyz2<0.0] *= -1.0
        T_inv[np.broadcast_to(np.isclose(qxy,0.0,rtol=0.0,atol=1.0e-12),T_inv.shape)] = 0.0
        cu = np.block([T_inv, np.where(xyz3[...,2:3]<0.0,-np.ones_like(xyz3[...,2:3]),np.ones_like(xyz3[...,2:3])) \
                              * rs/np.sqrt(6.0/np.pi),
                      ])/ sc

    cu[np.isclose(np.sum(np.abs(ho),axis=-1),0.0,rtol=0.0,atol=1.0e-16)] = 0.0
    cu = np.take_along_axis(cu,_get_pyramid_order(ho,'backward'),-1)

    return cu

#---------- Cubochoric ----------

def cu2qu(cu):
    """Cubochoric vector to quaternion."""
    return ho2qu(cu2ho(cu))


def cu2om(cu):
    """Cubochoric vector to rotation matrix."""
    return ho2om(cu2ho(cu))


def cu2eu(cu):
    """Cubochoric vector to Bunge-Euler angles."""
    return ho2eu(cu2ho(cu))


def cu2ax(cu):
    """Cubochoric vector to axis angle pair."""
    return ho2ax(cu2ho(cu))


def cu2ro(cu):
    """Cubochoric vector to Rodrigues-Frank vector."""
    return ho2ro(cu2ho(cu))


def cu2ho(cu):
    """
    Cubochoric vector to homochoric vector.

    References
    ----------
    D. Roşca et al., Modelling and Simulation in Materials Science and Engineering 22:075013, 2014
    https://doi.org/10.1088/0965-0393/22/7/075013

    """
    with np.errstate(invalid='ignore',divide='ignore'):
        # get pyramide and scale by grid parameter ratio
        XYZ = np.take_along_axis(cu,_get_pyramid_order(cu,'forward'),-1) * sc
        order = np.abs(XYZ[...,1:2]) <= np.abs(XYZ[...,0:1])
        q = np.pi/12.0 * np.where(order,XYZ[...,1:2],XYZ[...,0:1]) \
                       / np.where(order,XYZ[...,0:1],XYZ[...,1:2])
        c = np.cos(q)
        s = np.sin(q)
        q = R1*2.0**0.25/beta/ np.sqrt(np.sqrt(2.0)-c) \
          * np.where(order,XYZ[...,0:1],XYZ[...,1:2])

        T = np.block([ (np.sqrt(2.0)*c - 1.0), np.sqrt(2.0) * s]) * q

        # transform to sphere grid (inverse Lambert)
        c = np.sum(T**2,axis=-1,keepdims=True)
        s = c *         np.pi/24.0 /XYZ[...,2:3]**2
        c = c * np.sqrt(np.pi/24.0)/XYZ[...,2:3]
        q = np.sqrt( 1.0 - s)

        ho = np.where(np.isclose(np.sum(np.abs(XYZ[...,0:2]),axis=-1,keepdims=True),0.0,rtol=0.0,atol=1.0e-16),
                      np.block([np.zeros_like(XYZ[...,0:2]),np.sqrt(6.0/np.pi) *XYZ[...,2:3]]),
                      np.block([np.where(order,T[...,0:1],T[...,1:2])*q,
                                np.where(order,T[...,1:2],T[...,0:1])*q,
                                np.sqrt(6.0/np.pi) * XYZ[...,2:3] - c])
                      )

    ho[np.isclose(np.sum(np.abs(cu),axis=-1),0.0,rtol=0.0,atol=1.0e-16)] = 0.0
    ho = np.take_along_axis(ho,_get_pyramid_order(cu,'backward'),-1)

    return ho



def _get_pyramid_order(xyz,direction=None):
    """
    Get order of the coordinates.

    Depending on the pyramid in which the point is located, the order need to be adjusted.

    Parameters
    ----------
    xyz : numpy.ndarray
       coordinates of a point on a uniform refinable grid on a ball or
       in a uniform refinable cubical grid.

    References
    ----------
    D. Roşca et al., Modelling and Simulation in Materials Science and Engineering 22:075013, 2014
    https://doi.org/10.1088/0965-0393/22/7/075013

    """
    order = {'forward': np.array([[0,1,2],[1,2,0],[2,0,1]]),
             'backward':np.array([[0,1,2],[2,0,1],[1,2,0]])}

    p = np.where(np.maximum(np.abs(xyz[...,0]),np.abs(xyz[...,1])) <= np.abs(xyz[...,2]),0,
                 np.where(np.maximum(np.abs(xyz[...,1]),np.abs(xyz[...,2])) <= np.abs(xyz[...,0]),1,2))

    return order[direction][p]



from itertools import chain
from setuptools import setup, find_packages

from pyebsdindex import (
    __author__, __author_email__, __credits__, __description__, __name__, __version__
)


# Projects with optional features for building the documentation and running
# tests. From setuptools:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies
extra_feature_requirements = {
    "doc": [
        "nbsphinx                       >= 0.7",
        "numpydoc",
        "pydata-sphinx-theme",
        "sphinx                         >= 3.0.2",
        "sphinx-codeautolink[ipython]",
        "sphinx-copybutton              >= 0.2.5",
        "sphinx-design",
        "sphinx-gallery",
    ],
    "tests": [
        "coverage                       >= 5.0",
        "pytest                         >= 5.4",
        "pytest-cov                     >= 2.8.1",
    ],
    "gpu": [
        "pyopencl",
    ],
    "parallel": [
        "ray[default]                   >= 1.13",
        "pydantic                       < 2",
    ]
}
# Create a development installation "dev" including "doc" and "tests"
# projects
extra_feature_requirements["dev"] = list(
    chain(*list(extra_feature_requirements.values()))
)
# Create a user installation "all" including "gpu" and "parallel"
runtime_extras_require = {}
for x, packages in extra_feature_requirements.items():
    if x not in ["doc", "tests"]:
        runtime_extras_require[x] = packages
extra_feature_requirements["all"] = list(chain(*list(runtime_extras_require.values())))


setup(
    # Package description
    name=__name__,
    version=__version__,
    license="Custom",
    python_requires=">=3.7",
    description=__description__,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=[
        "EBSD",
        "electron backscatter diffraction",
        "HI",
        "Radon indexing",
        "NLPAR",
    ],
    zip_safe=True,
    # Contact
    author=__credits__,
    download_url="https://pypi.python.org/pypi/pyebsdindex",
    maintainer=__author__,
    maintainer_email=__author_email__,
    project_urls={
        "Bug Tracker": "https://github.com/USNavalResearchLaboratory/PyEBSDIndex/issues",
        "Documentation": "https://pyebsdindex.readthedocs.io",
        "Source Code": "https://github.com/USNavalResearchLaboratory/PyEBSDIndex",
    },
    url="https://pyebsdindex.readthedocs.io",
    # Dependencies
    extras_require=extra_feature_requirements,
    install_requires=[
        "h5py",
        "matplotlib",
        "numpy",
        "numba>=0.53",
        "scipy",
    ],
    # Files to include when distributing package (see also MANIFEST.in)
    packages=find_packages(),
    package_dir={"pyebsdindex": "pyebsdindex"},
    include_package_data=True,
)
